/****************************************************************************
 * <Novell-copyright>
 * Copyright (c) 2005 Novell, Inc. All Rights Reserved.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2 of the GNU General Public License
 * as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, contact Novell, Inc.
 * 
 * To contact Novell about this file by physical or electronic mail, you 
 * may find current contact information at www.novell.com.
 * </Novell-copyright>
****************************************************************************/

#ifndef CONTACT_H
#define CONTACT_H

#define NMAP_ADBK_VERSION       1L
#define NMAP_ADBK_PERSONAL      "PERSONAL"
#define NMAP_ADBK_SYSTEM        "SYSTEM"
#define NMAP_ADBK_EXT           ".abk"
#define NMAP_ADBK_INDEX_EXT     ".dba"

typedef enum _ContactFlags {
    ContactDeleted = (1 << 0),
    ContactPurged = (1 << 1)
} ContactFlags;

typedef enum _ContactType {
    ContactUnknown = 0,
    ContactPersonal = (1 << 0),
    ContactSystem = (1 << 1),
    ContactPersonalShared = (1 << 2),
    ContactReadOnly = (1 << 3)
} ContactType;

typedef enum _ContacParse {
    ContactOk = 0,
    ContactVersion = (1 << 0),
    ContactFN = (1 << 1),
    ContactN = (1 << 2)
} ContactParse;

typedef struct _ContactControlBlock {
    ContactType type;
    unsigned long newCount;
} ContactControlBlock;

typedef struct _ContactControl {
    unsigned long id;
    unsigned long flags;
    unsigned long version;
    unsigned long storeId;
    unsigned long storeSize;
    unsigned long storeCount;
    double modified;
    unsigned long fileSize;
} ContactControl;

typedef struct _ContactInfo {
    unsigned long id;
    unsigned long flags;
    unsigned char guid[NMAP_GUID_LENGTH];
    unsigned long record;
    unsigned long start;
    unsigned long fn;
    unsigned long n;
    unsigned long size;
    double indexed;
    double modified;
} ContactInfo;

typedef struct _ContactIndex {
    unsigned long size;
    unsigned long next;
    ContactInfo **pList;
} ContactIndex;

typedef struct _ContactSQL {
    sqlite3_stmt *begin;
    sqlite3_stmt *commit;
    sqlite3_stmt *abort;
    struct {
        /* select */
        sqlite3_stmt *read;
        /* insert */
        sqlite3_stmt *insertInit;
        /* update */
        sqlite3_stmt *update;
        /* drop */
        sqlite3_stmt *drop;
    } control;
    struct {
        /* select */
        sqlite3_stmt *readAll;
        sqlite3_stmt *modify;
        /* insert */
        sqlite3_stmt *insert;
        /* update */
        sqlite3_stmt *updateFlags;
        /* drop */
        sqlite3_stmt *drop;
        sqlite3_stmt *tableCreate;
    } index;
} ContactSQL;

typedef struct _NmapAddressBook {
    unsigned char name[XPL_MAX_PATH + 1];
    sqlite3 *handle;
    FILE *hFile;
    NLockStruct *lock;

    ContactControlBlock cb;
    ContactControl control;
    ContactIndex index;
    ContactSQL sql;

} NmapAddressBook;

/*
** vim:ts=4:sw=4:sts=4:et
*/

#endif
