/****************************************************************************
 * <Novell-copyright>
 * Copyright (c) 2001 Novell, Inc. All Rights Reserved.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2 of the GNU General Public License
 * as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, contact Novell, Inc.
 * 
 * To contact Novell about this file by physical or electronic mail, you 
 * may find current contact information at www.novell.com.
 * </Novell-copyright>
 ****************************************************************************/

#include "unix-mgr.h"

#include <config.h>
#if defined(SUSE) || defined(LINUX) || defined(S390RH) || defined(REDHAT) || defined(SOLARIS) || defined(MACOSX)
#include <sys/time.h>
#include <sys/types.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <unistd.h>
#include <popt.h>

#ifdef __cplusplus
extern "C" {
#endif

#include <msgapi.h>
#include <hulautil.h>
#include <logger.h>

#ifdef __cplusplus
}                                      // extern "C"
#endif
static Manager *pManager;

int
main(int argc, char *argv[])
{
    int err;
    Manager manager;

    err = manager.Init(argc, argv);
    if (err == 0) {
        err = manager.Run();
    }
    return (err);
}

DMC::DMC()
{
    char *ptr;

    strcpy(m_Name, MSGSRV_NLM_DMC);
    memset(m_Access, 0, DMC_HASH_SIZE);
    m_hDirectory = NULL;
    m_Load = 0;
    sprintf(m_Filename, "%s/%s", XPL_DEFAULT_BIN_DIR, m_Name);
    ptr = m_Filename + strlen(XPL_DEFAULT_BIN_DIR) + 1;
    while (*ptr != '\0') {
        if (*ptr != '.') {
            *ptr = tolower(*ptr);
        } else {
            if (XplStrCaseCmp(ptr, ".NLM") == 0) {
                *ptr = '\0';
                break;
            }
        }
        ptr++;
    }
    m_Pid = 0;
    m_pConn = NULL;
}

DMC::~DMC()
{
    if (m_pConn != NULL)
    {
        ConnFree(m_pConn);
        m_pConn = NULL;
    }
}

int
DMC::Init(MDBHandle hDirectory)
{
    int err = 1;

    m_hDirectory = hDirectory;
    if (m_hDirectory != NULL) {
        MDBValueStruct *pConfig = NULL;

        pConfig = MDBCreateValueStruct(m_hDirectory, NULL);
        if (pConfig != NULL) {
            MDBRead((unsigned char *) MSGSRV_ROOT,
                    (unsigned char *) MSGSRV_A_ACL, pConfig);
            if (pConfig->Used) {
                bool bResult = HashCredential(
                        MsgGetServerDN(NULL),
                        pConfig->Value[0],
                        m_Access);
                if (bResult) {
                    err = 0;
                }
            }
            MDBDestroyValueStruct(pConfig);
        }
        if (err == 0) {
            m_pConn = ConnAlloc(TRUE);
            if (m_pConn != NULL) {
                memset(&m_pConn->socketAddress, 0, sizeof(struct sockaddr_in));
                m_pConn->socketAddress.sin_family = AF_INET;
                m_pConn->socketAddress.sin_addr.s_addr = inet_addr("127.0.0.1");
                m_pConn->socketAddress.sin_port = htons(DMC_MANAGER_PORT);
            } else {
                err = 1;
            }
        }
    }
    return (err);
}

int
DMC::Load()
{
    int err;
    pid_t pid;

    XplSetEffectiveUser(0);
    pid = fork();
    err = SetUnprivileged();
    if (err == 0) {
        if (pid == 0) {
            char *ptr;

            setpgid(0, 0);
            ptr = strrchr(m_Filename, '/');
            XplConsolePrintf("Loading %s\n", ptr + 1);
            execl(m_Filename, m_Filename, NULL);
        } else if (pid > 0) {
            XplDelay(100);
            m_Load = time(NULL);
            m_Pid = pid;
        } else {
            err = 1;
        }
    }
    return (err);
}

int
DMC::Connect()
{
    int err;

    if (m_pConn->socket == -1)
    {
        m_pConn->socket = IPsocket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
    }
    if (m_pConn->socket > 0) {
        err = NonBlocking();
        if (err == 0) {
            for (int timeout = 1000; timeout > 0; timeout -= 100) {
                err = IPconnect(
                        m_pConn->socket,
                        (sockaddr *) & m_pConn->socketAddress,
                        sizeof(struct sockaddr_in));
                if (err > -1) {
                    err = Blocking();
                    break;
                } else {
                    XplDelay(100);
                }
            }
        }
    } else {
        err = 1;
    }
    return (err);
}

int
DMC::Authenticate()
{
    int err;
    char response[CONN_BUFSIZE + 1];

    err = ReadAnswer(response, CONN_BUFSIZE);
    if (err == 1000) {
        char *ptr = strchr(response, '\"');
        if (ptr != NULL) {
            char *pSalt = ++ptr;
            ptr = strchr(ptr, '\"');
            if (ptr != NULL) {
                MD5_CTX md5Context;
                unsigned char digest[16];
                unsigned char message[33];
                int indx;

                *ptr = '\0';
                MD5_Init(&md5Context);
                MD5_Update(&md5Context, pSalt, strlen(pSalt));
                MD5_Update(&md5Context, m_Access, DMC_HASH_SIZE);
                MD5_Final(digest, &md5Context);
                ConnWrite(m_pConn, "AUTH ", 5);
                for (indx = 0; indx < 16; indx++) {
                    sprintf(
                            (char *) (message + (indx * 2)),
                            "%02x",
                            digest[indx]);
                }
                ConnWrite(m_pConn, (char *) message, 32);
                ConnWrite(m_pConn, "\r\n", 2);
                err = ConnFlush(m_pConn);
                if (err = 39) {
                    err = ReadAnswer();
                    if (err == 1000) {
                        err = 0;
                    } else {
                        err = 1;
                    }
                } else {
                    err = 1;
                }
            }
        } else {
            err = 1;
        }
    } else {
        err = 1;
    }
    return (err);
}

int
DMC::Start()
{
    int err;

    ConnWrite(m_pConn, "START\r\n", 7);
    ConnFlush(m_pConn);
    err = ReadAnswer();
    if (err == 1000) {
        err = 0;
    } else {
        err = 1;
    }
    return (err);
}

int
DMC::Shutdown()
{
    int err;
    int errRt = 1;

    ConnWrite(m_pConn, "SHUTDOWN\r\n", 10);
    ConnFlush(m_pConn);
    for (int indx = 30; indx > 0; indx--) {
        err = ReadAnswer();
        if (err == 2004) {
            errRt = 0;
            break;
        } else {
            XplDelay(1000);
        }
    }
    return (errRt);
}

int
DMC::Quit()
{
    ConnWrite(m_pConn, "QUIT\r\n", 6);
    ConnFlush(m_pConn);
    return (0);
}

int
DMC::Close()
{
    CONN_TCP_CLOSE(m_pConn);
    return (0);
}

int
DMC::Kill()
{
    int err;

    XplConsolePrintf("  killing DMC\r\n");
    XplSetEffectiveUser(0);
    kill(m_Pid, SIGKILL);
    err = SetUnprivileged();
    return (err);
}

int
DMC::Blocking()
{
    int err = 1;
    int sockOpts;

    sockOpts = fcntl(m_pConn->socket, F_GETFL, NULL);
    if (sockOpts > -1) {
        sockOpts &= ~O_NONBLOCK;
        err = fcntl(m_pConn->socket, F_SETFL, sockOpts);
        if (err > -1) {
            err = 0;
        } else {
            err = 1;
        }
    }
    return (err);
}

int
DMC::NonBlocking()
{
    int err = 1;
    int sockOpts;

    sockOpts = fcntl(m_pConn->socket, F_GETFL, NULL);
    if (sockOpts > -1) {
        sockOpts |= O_NONBLOCK;
        err = fcntl(m_pConn->socket, F_SETFL, sockOpts);
        if (err > -1) {
            err = 0;
        } else {
            err = 1;
        }
    }
    return (err);
}

int
DMC::ReadAnswer(
        char *pResponse /* = NULL */ ,
        int iLenResponse /* = 0 */ ,
        bool bResult /* = TRUE */ )
{
    int err = -1;
    char buf[CONN_BUFSIZE + 1];
    int len;

    if (pResponse == NULL) {
        pResponse = buf;
        iLenResponse = CONN_BUFSIZE;
    }
    len = ConnReadAnswer(m_pConn, pResponse, iLenResponse);
    if (len > 0) {
        char *ptr = pResponse + 4;

        if ((bResult) &&
                ((*ptr == ' ') ||
                    (*ptr == '-') ||
                    ((ptr = strchr(pResponse, ' ')) != NULL))) {
            int count;

            *ptr++ = '\0';
            err = atoi(pResponse);
            count = len - (ptr - pResponse);
            memmove(pResponse, ptr, count);
            pResponse[count] = '\0';
        }
    }
    return (err);
}

int
DMC::SetUnprivileged()
{
    const char *pUser;
    int err = 0;

    pUser = MsgGetUnprivilegedUser();
    if (pUser != NULL) {
        err = XplSetEffectiveUser(pUser);
        if (err < 0) {
            XplConsolePrintf(
                    "hulamanager: Could not drop to"
                        " unprivileged user '%s', exiting.\n",
                    pUser);
            err = 4;
        } else {
            err = 0;
        }
    }
    return (err);
}

Manager::Manager()
{
    m_eState = eInitializing;
    m_bDaemonize = false;
    m_pid = 0;
    memset(m_pidPath, 0, XPL_MAX_PATH + 1);
    m_hDirectory = NULL;
    m_reload = 5 * 60;
    m_signum = SIGTERM;
    m_pLogging = NULL;
    memset(m_tree, 0, MDB_MAX_OBJECT_CHARS + 1);
    memset(m_host, 0, MDB_MAX_OBJECT_CHARS + 1);
    memset(m_dn, 0, MDB_MAX_OBJECT_CHARS + 1);
    m_pRdn = NULL;
}

Manager::~Manager()
{
    m_pid = 0;
}

int
Manager::Init(int argc, char *argv[])
{
    int err;

    err = SetLimits();
    if (err != 0) {
        return err;
    }

    SetUnprivileged();

    err = GetOptions(argc, argv);
    if (err == 0) {
        err = CheckInstance();
        if (err == 0) {
            err = MemoryManagerOpen(cProductShortName);
            if (err > 0) {
                unsigned long
                        ulTime;
                MD5_CTX
                        md5Context;
                unsigned char
                        buf[256];

                IPInit();
                XplGetHighResolutionTime(ulTime);
                SSL_load_error_strings();
                SSL_library_init();
                XPLCryptoLockInit();
                srand(ulTime);
                MD5_Init(&md5Context);
                for (int indx = 0; indx < sizeof(buf);) {
                    buf[indx] = rand();
                    if (buf[indx] > 0) {
                        indx++;
                    }
                }
                MD5_Update(&md5Context, buf, sizeof(buf) - 1);
                MD5_Final(buf, &md5Context);
                RAND_seed(buf, 16);
                StartLogging();
                err = ConnStartup(15 * 60, FALSE);
                if (err != 0) {
                    err = MDBInit();
                    if (err != 0) {
                        m_hDirectory = MsgInit();
                        if (m_hDirectory != NULL) {
                            err = CheckConfiguration();
                            if (err == 0) {
                                err = dmc.Init(m_hDirectory);
                            }
                        } else {
                            XplConsolePrintf(
                                    "Hula messageing libary failed"
                                        " to initialize.\n\r");
                            err = 1;
                        }
                    } else {
                        XplConsolePrintf(
                                "Hula directory interface failed"
                                    " to initialize.\n\r");
                        err = 1;
                    }
                } else {
                    XplConsolePrintf(
                            "Hula connection management library"
                                " failed to initialize.\n\r");
                    err = 1;
                }
            } else {
                err = 1;
            }
        }
    }
    if (err == 0) {
        CONN_TRACE_INIT(MsgGetWorkDir(NULL), "manager");
    } else {
        XplConsolePrintf("Host not configured fo Hula.\n\r");
    }
    return (err);
}

int
Manager::Run()
{
    int err;

    err = SetUnprivileged();
    if (err == 0) {
        while (!IsDone() && !IsTerminate()) {
            if (IsRunning()) {
                err = Status();
            } else if (IsReloading()) {
                err = Reload();
            } else if (IsStopping()) {
                err = Stop();
            } else {
                err = Start();
            }
        }
    }

    return (err);
}

int
Manager::SetLimits()
{
    int err;
    int privErr;
    struct rlimit rlim;

    rlim.rlim_max = 20480;
    rlim.rlim_cur = 20480;
    err = setrlimit(RLIMIT_NOFILE, &rlim);

    if (err != 0) {
        XplConsolePrintf("couldn't set rlimit: %s\n", strerror(errno));
    }

    return (err);
}

int
Manager::SetUnprivileged()
{
    const char *pUser;
    int err = 0;

    pUser = MsgGetUnprivilegedUser();
    if (pUser != NULL) {
        err = XplSetEffectiveUser(pUser);
        if (err < 0) {
            XplConsolePrintf(
                    "hulamanager: Could not drop to"
                        " unprivileged user '%s', exiting.\n",
                    pUser);
            err = 4;
        } else {
            err = 0;
        }
    }
    return (err);
}

int
Manager::GetOptions(int argc, char *argv[])
{
    int err = 0;
    poptContext optContext;
    struct poptOption optTable[] = {
        { "retstart", 'r', POPT_ARG_VAL, &m_eState, eReloading, "Restart running Hula agents", NULL},
        { "status", '\0', POPT_ARG_VAL, &m_eState, eRunning, "Report on running Hula agents", NULL},
        { "stop", 's', POPT_ARG_VAL, &m_eState, eStopping, "Terminate running Hula agents", NULL},
        { "daemon", 'd', POPT_ARG_NONE, NULL, 'd', "Run as a daemon", NULL},
        POPT_AUTOHELP {NULL, 0, 0, NULL, 0, NULL}
    };

    optContext = poptGetContext(NULL, argc, (const char **) argv, optTable, 0);
    if (optContext != NULL) {
        int val;

        poptSetOtherOptionHelp(optContext, "[OPTION]");
        while ((val = poptGetNextOpt(optContext)) >= 0) {
            if (val == 'd') {
                m_bDaemonize = true;
            }
        }
        if (val < -1) {
            fprintf(stderr,
                    "%s: %s\n",
                    poptBadOption(optContext, POPT_BADOPTION_NOALIAS),
                    poptStrerror(val));
            err = 2;
        } else if (poptPeekArg(optContext) != NULL) {
            poptPrintUsage(optContext, stderr, 0);
            err = 2;
        }
        poptFreeContext(optContext);
    } else {
        err = 1;
    }
    return (err);
}

int
Manager::CheckInstance()
{
    int err = 0;
    char line[XPL_MAX_PATH + 1];
    FILE *hFile;

    m_pid = 0;
    strcpy(line, XPL_DEFAULT_WORK_DIR);
    XplMakeDir(line);
    sprintf(m_pidPath, "%s/hulamanager.pid", XPL_DEFAULT_WORK_DIR);
    hFile = fopen(m_pidPath, "r");
    if (hFile != NULL) {
        line[0] = '\0';
        fread(line, sizeof(char), sizeof(line), hFile);
        fclose(hFile);
        m_pid = strtol(line, NULL, 10);
        if (kill(m_pid, 0) != 0) {
            m_pid = 0;
            unlink(m_pidPath);
            err = 0;
        }
    }
    return (err);
}

int
Manager::Status()
{
    int err = 7;

    if (m_pid != 0) {
        XplConsolePrintf(
                "Hula manager is currently running, requesting status.\r\n");
        if (kill(m_pid, SIGUSR1) != 0) {
            err = 1;
        }
    } else {
        XplConsolePrintf("Hula manager is not currently running.\n\r");
    }
    m_eState = eDone;
    return (err);
}

int
Manager::Reload()
{
    int err = 7;

    if (m_pid != 0) {
        XplConsolePrintf(
                "Hula manager is currently running,"
                    " attempting to restart.\r\n");
        if (kill(m_pid, SIGHUP) != 0) {
            err = 1;
        }
    } else {
        XplConsolePrintf(
                "Hula manager is not currently running,"
                    " requesting status.\r\n");
    }
    m_eState = eDone;
    return (err);
}

int
Manager::Stop()
{
    int err = 7;

    if (m_pid != 0) {
        XplConsolePrintf(
                "Hula manager is currently running, attempting to stop.\r\n");
        if (kill(m_pid, SIGTERM) != 0) {
            err = 1;
        }
    } else {
        XplConsolePrintf(
               "Hula manager is not currently running, requesting status.\r\n");
    }
    m_eState = eDone;
    return (err);
}

int
Manager::Start()
{
    int err = 8;

    if (m_pid == 0) {
        err = Daemonize();
        if (err == 0) {
            err = WritePidFile();
            if (err == 0) {
                XplConsolePrintf(
                        "Loading Hula agents for server %s.\n\r",
                        m_pRdn);
                setsid();
                err = SetSignalHandler();
                err = StartAgents();
                if (err == 0) {
                    m_eState = eRunning;
                    while (IsRunning()) {
                        XplDelay(5000);
                    }
                    StopAgents();
                    err = 100;
                }
            } else {
                m_eState = eDone;
                err = 1;
            }
        } else {
            m_eState = eDone;
        }
    } else {
        XplConsolePrintf(
                "Hula manager is currently running, process id %d.\r\n",
                m_pid);
    }
    return (err);
}

int
Manager::StartLogging()
{
    int err = 0;

    m_pLogging = LoggerOpen("hulamanager");
    if (m_pLogging == NULL) {
        XplConsolePrintf(
                "Hula manager failed to acquire a logging handle.\r\n");
        err = 1;
    }
    return (err);
}

int
Manager::CheckConfiguration()
{
    int err;
    MDBValueStruct *pConfig = NULL;

    pConfig = MDBCreateValueStruct(m_hDirectory, NULL);
    if (pConfig != NULL) {
        BOOL bResult;

        m_tree[0] = '\\';
        bResult = MDBGetServerInfo(m_host, m_tree + 1, pConfig);
        if (bResult) {
            int indx;

            err = 0;
            indx = MDBReadDN(
                    m_host,
                    (unsigned char *) MSGSRV_A_HULA_MESSAGING_SERVER,
                    pConfig);
            if (indx != 0) {
                strcpy(m_dn, (char *) pConfig->Value[0]);
            } else {
                indx = MDBReadDN(
                        m_host,
                        (unsigned char *) MSGSRV_A_MESSAGING_SERVER,
                        pConfig);
                if (indx != 0) {
                    strcpy(m_dn, (char *) pConfig->Value[0]);
                    indx = MDBWriteDN(
                            m_host,
                            (unsigned char *) MSGSRV_A_HULA_MESSAGING_SERVER,
                            pConfig);
                    if (indx != 0) {
                        MDBClear(
                                m_host,
                                (unsigned char *) MSGSRV_A_MESSAGING_SERVER,
                                pConfig);
                    } else {
                        err = 1;
                        XplConsolePrintf(
                                "Hula messaging server not upgraded.\n\r");
                    }
                } else {
                    err = 1;
                    XplConsolePrintf(
                            "Hula messaging server not configured.\n\r");
                }
            }
            MDBFreeValues(pConfig);
            if (err == 0) {
                m_pRdn = strrchr(m_dn, '\\') + 1;
            }
            MDBFreeValues(pConfig);
        } else {
            XplConsolePrintf("Hula directory driver host not configured\n\r");
            err = 1;
        }
        MDBDestroyValueStruct(pConfig);
        pConfig = NULL;
    } else {
        XplConsolePrintf("Hula directory driver not responding\n\r");
        err = 1;
    }
    return (err);
}

int
Manager::Daemonize()
{
    int err = 0;

    if (m_bDaemonize) {
        pid_t pid;

        pid = fork();
        if (pid < 0) {
            err = 1;
        } else if (pid == 0) {
            int fd;

            setsid();
            chdir("/");
            for (fd = getdtablesize(); fd > -1; fd--) {
                close(fd);
            }
            fd = open("/dev/null", O_RDWR);
            dup(fd);
            dup(fd);
        } else {
            exit(0);
        }
    }
    return (err);
}

int
Manager::WritePidFile()
{
    int err = 0;
    FILE *hFile;

    hFile = fopen(m_pidPath, "w+");
    if (hFile != NULL) {
        fprintf(hFile, "%d", getpid());
        fclose(hFile);
    } else {
        err = 1;
        XplConsolePrintf("Hula manager could not create \"%s\"; error %d\n",
                m_pidPath, errno);
    }
    return (err);
}

int
Manager::SetSignalHandler()
{
    int err;
    int indx;
    struct sigaction
    act = { 0 };
    sigset_t sigSet;
    int block[] = {
        SIGUSR2,
        SIGALRM,
        SIGPROF,
        SIGVTALRM,
#ifdef SIGPOLL
        SIGPOLL,
#endif
#ifdef SIGSTKFLT
        SIGSTKFLT,
#endif
#ifdef SIGPWR
        SIGPWR,
#endif
#ifdef SIGUNUSED
        SIGUNUSED,
#endif
        0
    };
    int action[] = {
        SIGTERM,
        SIGINT,
        SIGUSR1,
        SIGHUP,
        SIGPIPE,
        SIGCHLD,
        0
    };

    pManager = this;
    sigemptyset(&sigSet);
    for (indx = 0; block[indx] != 0; indx++) {
        sigaddset(&sigSet, block[indx]);
    }
    pthread_sigmask(SIG_BLOCK, &sigSet, NULL);
    sigemptyset(&sigSet);
    for (indx = 0; action[indx] != 0; indx++) {
        sigaddset(&sigSet, action[indx]);
    }
    pthread_sigmask(SIG_BLOCK, &sigSet, NULL);
    act.sa_sigaction = SignalHandler;
    for (indx = 0; action[indx] != 0; indx++) {
        (void) sigaction(action[indx], &act, NULL);
    }
    pthread_sigmask(SIG_UNBLOCK, &sigSet, NULL);
    return (0);
}

int
Manager::StartAgents()
{
    int err;

    err = dmc.Load();
    if (err == 0) {
        err = dmc.Connect();
        if (err == 0) {
            err = dmc.Authenticate();
            if (err == 0) {
                err = dmc.Start();
            }
            dmc.Quit();
            dmc.Close();
        }
    }
    return (err);
}

int
Manager::StopAgents()
{
    int err;

    err = dmc.Connect();
    if (err == 0) {
        bool bKill;

        err = dmc.Authenticate();
        if (err == 0) {
            bKill = dmc.Shutdown();
        }
        dmc.Quit();
        dmc.Close();
        if (bKill) {
            dmc.Kill();
            err = 0;
        }
    }
    return (err);
}

void
Manager::SignalHandler(int signum, siginfo_t *, void *)
{
    switch (signum) {
        case SIGHUP:
            pManager->m_eState = eReloading;
            break;
        case SIGINT:
        case SIGTERM:
            if (pManager->IsStopping()) {
                pManager->m_signum = SIGKILL;
            } else {
                pManager->m_eState = eTerminate;
            }
        case SIGCHLD:
        {
            pid_t pid;
            int stat;

            while ((pid = waitpid(-1, &stat, WNOHANG)) > 0) {
                pManager->m_eState = eDone;
            }
            break;
        }

            break;
        default:
            break;
    }
}

#endif                                 /*  defined(SUSE) || defined(LINUX) || defined(S390RH) || defined(REDHAT) || defined(SOLARIS) || defined(MACOSX)    */
