/****************************************************************************
 * <Novell-copyright>
 * Copyright (c) 2001 Novell, Inc. All Rights Reserved.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2 of the GNU General Public License
 * as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, contact Novell, Inc.
 * 
 * To contact Novell about this file by physical or electronic mail, you 
 * may find current contact information at www.novell.com.
 * </Novell-copyright>
 ****************************************************************************/
#ifndef UNIX_MGR_H
#define UNIX_MGR_H

#include <config.h>
#include <sys/types.h>

#ifdef __cplusplus
extern "C" {
#endif

#include <xpl.h>
#include <mdb.h>
#include <management.h>

#ifdef __cplusplus
}                                      // extern "C"
#endif

#ifdef __cplusplus
const unsigned char cProductShortName[] = "Hula";

/*!
** \class   DMC
** \brief   <description>
**
** \note    <note>
*/
class DMC {
    /*
    ** methods
    */
    public:
        DMC();
        ~DMC();

        int Init(MDBHandle hDirectory);
        int Load();
        int Connect();
        int Authenticate();
        int Start();
        int Shutdown();
        int Quit();
        int Close();
        int Kill();
    protected:
    private:
        int Blocking();
        int NonBlocking();
        int ReadAnswer(
                char *pResponse = NULL,
                int iLenResponse = 0,
                bool bResult = TRUE);
        int SetUnprivileged();

        /*
        ** Data
        */
    public:
    protected:
    private:
        char m_Name[XPL_MAX_PATH + 1];
        unsigned char m_Access[DMC_HASH_SIZE];
        MDBHandle m_hDirectory;
        time_t m_Load;
        char m_Filename[XPL_MAX_PATH + 1];
        pid_t m_Pid;
        Connection *m_pConn;
};

/*!
** \class   Manager
** \brief   <description>
**
** \note    <note>
*/
class Manager {
    enum States {
        eStarting = 0,
        eInitializing,
        eLoading,
        eRunning,
        eReloading,
        eUnloading,
        eTerminate,
        eStopping,
        eDone,
        eMaxStates
    };

    /*
    ** methods
    */
    public:
        Manager();
        ~Manager();

        int Init(int argc, char *argv[]);
        int Run();
        protected:
        private:
        int SetLimits();
        int SetUnprivileged();
        int GetOptions(int argc, char *argv[]);
        int CheckInstance();

        inline bool IsDone() { return (m_eState == eDone); }
        inline bool IsReloading() { return (m_eState == eReloading); }
        inline bool IsStopping() { return (m_eState == eStopping); }
        inline bool IsTerminate() { return (m_eState == eTerminate); }
        inline bool IsRunning() { return (m_eState == eRunning); }

        int Status();
        int Reload();
        int Stop();
        int Start();
        int StartLogging();
        int CheckConfiguration();
        int Daemonize();
        int WritePidFile();
        int SetSignalHandler();
        int StartAgents();
        int StopAgents();

        static void SignalHandler(int sig, siginfo_t * pInfo, void *context);
    /*
    ** Data
    */
    public:
    protected:
    private:
        DMC dmc;

        States m_eState;
        bool m_bDaemonize;
        pid_t m_pid;
        char m_pidPath[XPL_MAX_PATH + 1];
        MDBHandle m_hDirectory;
        int m_reload;
        int m_signum;
        void *m_pLogging;
        unsigned char m_tree[MDB_MAX_OBJECT_CHARS + 1];
        unsigned char m_host[MDB_MAX_OBJECT_CHARS + 1];
        char m_dn[MDB_MAX_OBJECT_CHARS + 1];
        char *m_pRdn;
};

#endif

#endif
