using System;
using System.Xml;
using System.IO;
using System.Collections;
using System.Text;
using Hula.ICal;
using Hula.Util;

namespace Hula.Cal {

public class VCardFilter : Filter {

    public bool allProp;
    public Hashtable properties = null;

    public VCardFilter() {
	allProp = true;
    }

    public VCardFilter(XmlNode node) {
	if (node != null && node.FirstChild != null) {
	    node = node.FirstChild;

	    foreach (XmlNode child in node) {
		string name = null;
		switch (child.LocalName) {
		case "prop" :
		    name = child.Attributes["name"].Value;
		    if (properties == null) {
			properties = new Hashtable();
		    }
		    properties[name] = true;
		    break;
		case "allprop" :
		    allProp = true;
		    break;
		}
	    }
	} else {
	    allProp = true;
	}
    }
	
    private static string ReadFoldedLine(TextReader reader) {
	string line = reader.ReadLine();
	if (line == null) {
	    return null;
	}
	line = line.Replace("\\n", "\n");
	int c = reader.Peek();
	if (c == ' ') {
	    StringBuilder builder = new StringBuilder();
	    builder.Append(line);
	    do {
		string newLine = reader.ReadLine();
		if (newLine == null) {
		    break;
		}
		newLine = newLine.Replace("\\n", "\n");
		    
		builder.Append(newLine.Substring(1));
		c = reader.Peek();
	    } while (c == ' ');
	    return builder.ToString();		
	} else {
	    return line;
	}
    }

    public override void Run(TextReader reader, TextWriter writer) {
	if (properties != null) {
	    base.Run(reader, writer);
	}

	string line = ReadFoldedLine(reader);
	while (line != null) {
	    string[] split;
	    if (line.IndexOf(':') < line.IndexOf(';')) {
		split = line.Split(new char[] {':'}, 2);
	    } else {
		split = line.Split(new char[] {':'}, 2);
	    }

	    if (split[0] == "BEGIN" || properties.Contains(split[0])) {
		writer.WriteLine(line);
	    }
	    line = ReadFoldedLine(reader);
	}
    }
}
}
