using System;
using System.Xml;
using System.Web;
using System.Text;
using System.Collections;
using System.IO;
using Lucene.Net.Search;
using Lucene.Net.Index;
using Lucene.Net.Documents;
using Hula.Store;
using Hula.Cal;
using Hula.Indexing;
using Hula.Util;

namespace Hula.Dav {

public class DavMultigetReport : DavReport {
    protected ArrayList hrefs = new ArrayList();
    private bool isCollection;


    public DavMultigetReport(HttpContext context) : base (context) {
    }

    private bool IsFolder(LuceneIndex idx, IStore store, string path) {
	BooleanQuery bq = new BooleanQuery();
	TermQuery tq = new TermQuery(new Term("hula.owner", store.Owner));
	bq.Add(tq, true, false);

	tq = new TermQuery(new Term("hula.store", store.Name));
	bq.Add(tq, true, false);

	tq = new TermQuery(new Term("hula.davpath", path));
	bq.Add(tq, true, false);

	Hits hits = idx.Searcher.Search(bq);
	if (hits != null && hits.Length () > 0 &&
	    hits.Doc(0).GetField("hula.mime-type").StringValue() == "application/x-directory") {
	    return true;
	}
        Console.Error.WriteLine ("owner: {0} name: {1} path: {2} hits: {3} type: {4}",
                                 store.Owner, store.Name, path,
                                 hits == null ? -1 : hits.Length (),
                                 hits == null ? ">>>null<<<" : hits.Length () <= 0 ? ">>>none<<<" : 
                                 hits.Doc(0).GetField("hula.mime-type").StringValue ());
	return false;
    } 
    
    public override void Setup(LuceneIndex idx, 
			       IStore store,
			       string path,
			       int depth,
			       XmlNode node) {
	isCollection = IsFolder(idx, store, path);	
	base.Setup(idx, store, path, depth, node);
    }

    protected override void HandleNode(XmlNode node) {
	switch(node.LocalName) {
	case "href" :
	    ParseHRef(node, store, path, isCollection);	    
	    break;
	default :
	    base.HandleNode(node);
	    break;
	}
    }

    public override void Run() {
	if (hrefs.Count == 0) {
	    throw new HttpException (405, "At least one href is required");
	}

	BeginResponse();

	/* We actually want a <response> for every <href> the client
	 * provided, not just a general query
	 */
	foreach (string href in hrefs) {
	    TermQuery q = new TermQuery (new Term ("hula.davhref", href));
	    Hits hits = idx.Searcher.Search(q);
	    if (hits.Length () == 0) {
		Response(store, null, "404 Not found");
	    } else {
		Document doc = hits.Doc(0);
		Response(store, doc);
	    }
	}

	EndResponse();
    }

    public string RemovePrefix(string str, string prefix) {
	if (str.StartsWith(prefix)) {
	    return str.Substring(prefix.Length);
	} else {
	    return null;
	}
    }

    public void ParseHRef (XmlNode node, IStore store, string path, bool isCollection) {
	string uri = node.InnerText;
	string href;
	string hrefOwner;
	string hrefStore;
	string hrefPath;

	if (hrefs.Count > 1 && !isCollection) {
	    throw new HttpException (405, "Multiple hrefs supplied on non-collection");
	}

	href = RemovePrefix (uri, appUrl);
	if (href == null) {
	    href = RemovePrefix (uri, "/dav");
	}
	if (href == null) {
	    throw new HttpException (405, "Request URI does not contain href");
	}

	DavService.DecomposeUrl(href, out hrefOwner,
				out hrefStore,
				out hrefPath);

	if (hrefOwner != store.Owner ||
	    hrefStore != store.Name ||
	    !hrefPath.StartsWith (path)) {
	    throw new HttpException (405, "Request URI does not contain href");
	}
	
	hrefs.Add (href);
    }
}
}
