using System;
using System.Collections;
using System.Xml;
using System.IO;
using Hula.Store;
using Lucene.Net.Documents;

namespace Hula.Dav {
public class PropertyFormatter : ResultFormatter {
    public bool AllProps = true;
    private IDictionary requestedProperties = null;

    private static Hashtable myNamespaces = new Hashtable();
    private static Hashtable ignoredProperties = new Hashtable();

    static PropertyFormatter() {
	myNamespaces["urn:ietf:params:xml:ns:carddav"] = "carddav";
	myNamespaces["urn:ietf:params:xml:ns:caldav"] = "C";
	myNamespaces["http://hula-project.org/Xml_Namespace/Hula"] = "H";
	myNamespaces["DAV:"] = "D";
	
	ignoredProperties["H:nmap_resource"] = true;
	ignoredProperties["H:nmap_modified"] = true;
	ignoredProperties["H:nmap_type"] = true;
	ignoredProperties["H:hula_store"] = true;
	ignoredProperties["H:hula_folder"] = true;
	ignoredProperties["H:hula_davfilename"] = true;
    }

    public void ParseFilter(XmlNode node) {
	AllProps = false;
	requestedProperties = new Hashtable();
	foreach (XmlNode child in node) {
	    if (myNamespaces.Contains(child.NamespaceURI)) {
		requestedProperties[(string)myNamespaces[child.NamespaceURI] + ":" + child.LocalName] = true;
	    }		
	}
    }

    public void ParsePropFind(XmlDocument doc) {
	foreach (XmlNode node in doc.DocumentElement) {
	    if (node.LocalName == "allprop") {
		AllProps = true;
	    } else if (node.LocalName == "prop") {
		ParseFilter(node);
	    }
	}
    }

    private void FormatProperties(TextWriter writer,
				  string resourceType,
				  IDictionary properties,
				  string status) {
	writer.WriteLine(" <D:propstat>");
	writer.WriteLine("  <D:prop>");

	bool includeType;
	IEnumerable propNames;
	if (AllProps) {
	    propNames = properties.Keys;
	    includeType = true;
	} else {
	    propNames = requestedProperties.Keys;
	    includeType = false;
	}

	foreach (string prop in propNames) {
	    if (AllProps && ignoredProperties.Contains(prop)) {
		continue;
	    }
	    if (prop == "D:resourcetype") {
		includeType = true;
	    } else {
		writer.WriteLine("<{0}>{1}</{0}>", prop, properties[prop]);
	    }
	}

	if (includeType) {
	    writer.Write("<D:resourcetype>");
	    writer.Write(resourceType);
	    writer.WriteLine("</D:resourcetype>");
	}

	writer.WriteLine("  </D:prop>");	
	writer.WriteLine("  <D:status>HTTP/1.1 {0}</D:status>", status);
	writer.WriteLine(" </D:propstat>");
    }

    public override void Format(TextWriter writer, IStore store, string status) {
	Hashtable props = new Hashtable();
	
	FormatProperties(writer,
			 "<D:collection/><H:store/>",
			 props,
			 status);
    }

    public override void Format(TextWriter writer, IStore store, Document doc, string status) {
	string guid = doc.GetField("guid").StringValue();
	IDictionary storeProps = store.GetProperties(guid);
	Hashtable formatProps = new Hashtable();
	string resourceType = "";
	// Rewrite properties to dav properties
	foreach (string key in storeProps.Keys) {
	    bool keepProp = true;
	    switch (key) {
	    case "hula.mime-type" :
		if ((string)storeProps["hula.mime-type"] == "application/x-directory") {
		    if (store.Name == "calendar") {
			resourceType = "<D:collection/><C:calendar/>";
		    } else {
			resourceType = "<D:collection/>";
		    }
		}
		formatProps["D:getcontenttype"] = resourceType;
		keepProp = false;
		break;
	    case "hula.etag" :
		formatProps["D:getetag"] = storeProps[key];
		keepProp = false;
		break;
	    case "hula.length" :
		formatProps["D:getcontentlength"] = storeProps[key];
		keepProp = false;
		break;
	    }
	    
	    if (keepProp) {
		string newProp = "H:" + key.Replace(".", "_");
		formatProps[newProp] = storeProps[key];
	    }
	}

	FormatProperties(writer, resourceType, formatProps, status);
    }
}
}
