using System;
using System.IO;
using System.Text;
using System.Collections;
using Lucene.Net;
using Lucene.Net.Analysis.Standard;
using Lucene.Net.Documents;
using Lucene.Net.Index;
using Lucene.Net.Store;
using Hula.ICal;
using Hula.Store;

using log4net;

namespace Hula.Indexing {
public class MailIndexer : Indexer {
    private static readonly ILog log = LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);

    private static Hashtable indexHeaders = new Hashtable();
    private static Hashtable headerScores = new Hashtable();

    static MailIndexer() {
        indexHeaders.Add("subject", true);
        indexHeaders.Add("to", true);
        indexHeaders.Add("from", true);
        indexHeaders.Add("date", true);

        headerScores.Add("subject", (float)2.0);
        headerScores.Add("to", null);
        headerScores.Add("from", null);
        headerScores.Add("date", null);
    }

    public MailIndexer() : base() {
    }

    public override void IndexDocument(LuceneIndex idx,
				       IStore store,
				       string guid) {
	TextReader reader;
	IDictionary properties;
        IDictionary contentFields = new Hashtable();

	GetDocument(store, guid, out reader, out properties);

	using(reader) {
            Document doc = new Document();

            char[] headerSplit = {':'};

            while(true) {
                string line = reader.ReadLine();

                if (line == null || line.Equals("")) {
                    break;
                }

                string[] fields = line.Split(headerSplit, 2);

                if (fields == null || fields.Length < 2) {
                    continue;  /* next line */
                }

                string header = fields[0].ToLower();

                if (indexHeaders.Contains(header)) {
                    string content = fields[1].Trim();
                    contentFields.Add(header, content);
                }
            }

            AddProperties(doc, properties);
            AddStandardFields(doc, store, guid, properties);
            AddContentFields(doc, contentFields, headerScores);
            
            string body = reader.ReadToEnd();

            if (body != null) {
                doc.Add(Field.UnStored("body", body));
                doc.Add(Field.UnStored("everything", body));
            }

            if (doc.Get("subject") != null) {
                doc.Add(Field.Text("summary", doc.Get("subject")));
            }
            
            idx.Writer.AddDocument(doc);
	}
    }
}
}
