/****************************************************************************
 *
 * Copyright (c) 2001-2002 Novell, Inc.
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2.1 of the GNU Lesser General Public
 * License as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, contact Novell, Inc.
 *
 * To contact Novell about this file by physical or electronic mail,
 * you may find current contact information at www.novell.com
 *
 ****************************************************************************/
using System;
using System.Runtime.InteropServices;
using System.Text;

using MDBHandle = System.IntPtr;

namespace Hula.Mdb {

	public class MdbUserObject : MdbObject {
	
		protected string surname;
	
		public string Surname {
			get {
				return surname;
			}
		}
	
		public MdbUserObject(string dn, string surname) : base(dn, C_USER) {
			this.surname = surname;
		}
		
		public static MdbUserObject CreateUserObject(MDBHandle handle, 
				string dn, string surname) {
			MdbUserObject uobj = new MdbUserObject(dn, surname);
		
			uobj.v = MDBCreateValueStruct(handle, uobj.dn);
			if (uobj.v == NULL) {
				return null;
			}
			
			IntPtr attribute = MDBShareContext(uobj.v);
			if (attribute == NULL) {
				MDBDestroyValueStruct(uobj.v);
				return null;
			}
			
			IntPtr data = MDBShareContext(uobj.v);
			if (data == NULL) {
				MDBDestroyValueStruct(attribute);
				MDBDestroyValueStruct(uobj.v);
				return null;
			}
			
			/* eDir requires surname for creation */
			if (!MDBAddStringAttribute(A_SURNAME, uobj.surname, attribute, data)) {
				MDBDestroyValueStruct(data);
				MDBDestroyValueStruct(attribute);
				MDBDestroyValueStruct(uobj.v);
				return null;
			}
			
			if (!MDBCreateObject(uobj.dn, uobj.Class, attribute, data, uobj.v)) {
				MDBDestroyValueStruct(data);
				MDBDestroyValueStruct(attribute);
				MDBDestroyValueStruct(uobj.v);
				return null;
			}
			
			MDBDestroyValueStruct(data);
			MDBDestroyValueStruct(attribute);
			
			return uobj;
		}
	
		public override string ToString() {
			string outstr = "";
			
			outstr += "MDB User Info\n";
			outstr += "-------------\n";
			outstr += "DN: " + Dn + "\n";
			outstr += "Surname: " + Surname + "\n";
			
			return outstr;
		}
	}
}
