/****************************************************************************
 *
 * Copyright (c) 2005 Novell, Inc.
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2.1 of the GNU Lesser General Public
 * License as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, contact Novell, Inc.
 *
 * To contact Novell about this file by physical or electronic mail,
 * you may find current contact information at www.novell.com
 *
 ****************************************************************************/

using System;
using System.Collections;
using System.Text.RegularExpressions;

using log4net;

namespace Hula.Nmap
{
public struct MessageInfo
{
    public int seq;
    public long msgLength;
    public long headLength;
    public long bodyLength;
    public string flags;
    public string wuid;
    public string uid;
    public string mboxId;
    public string date;
}

public struct MimeInfo
{
    public string type;
    public string subtype;
    public string charset;
    public string encoding;
    public string name;
    public long start;
    public long len;
    public long headLen;
    public int lines;
}

public class MboxCommands : NmapCommands
{
    private static readonly ILog log = LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);

    private static Regex mimeRe = null;

    static MboxCommands ()
    {
        mimeRe = new Regex("(?<=\")[^\"]*(?=\")|[^ \"]+");
    }

    public MboxCommands(NmapConnection nconn) : base(nconn)
    {
    }

    private MessageInfo GetMessageInfo(string info)
    {
        string[] vals = info.Split(' ');

        MessageInfo mi = new MessageInfo();
        mi.seq = Convert.ToInt32 (vals[0]);
        mi.msgLength = Convert.ToInt64 (vals[1]);
        mi.headLength = Convert.ToInt64 (vals[2]);
        mi.bodyLength = Convert.ToInt64 (vals[3]);
        mi.flags = vals[4];
        mi.wuid = vals[5];
        mi.uid = vals[6];
        mi.mboxId = vals[7];
        mi.date = vals[8];

        return mi;
    }

    private MimeInfo GetMimeInfo(string info)
    {
        MatchCollection matches = mimeRe.Matches(info);

        MimeInfo mi = new MimeInfo();
        mi.type = matches[0].Value.ToLower();
        mi.subtype = matches[1].Value.ToLower();
        mi.charset = matches[2].Value.ToLower();
        mi.encoding = matches[3].Value.ToLower();
        mi.name = matches[4].Value;
        mi.start = Convert.ToInt64 (matches[5].Value);
        mi.len = Convert.ToInt64 (matches[6].Value);
        mi.headLen = Convert.ToInt64 (matches[7].Value);
        mi.lines = Convert.ToInt32 (matches[8].Value);

        return mi;
    }

    public IList INFO()
    {
        nconn.WriteRequest("INFO");
        NmapResponse response = nconn.ReadLongReply ();

        if (response.data.Equals(""))
        {
            return new ArrayList();
        }

        string[] lines = response.data.Split('\n');
        IList ret = new ArrayList(lines.Length);

        for(int i=0; i<lines.Length; i++)
        {
            ret.Add(GetMessageInfo(lines[i]));
        }

        return ret;
    }

    public MessageInfo INFO(int seq)
    {
        nconn.WriteRequest("INFO {0}", seq);
        NmapResponse response = nconn.ReadFullReply ();
        return GetMessageInfo(response.text);
    }

    public NmapStreamReader LIST(int seq)
    {
        nconn.WriteRequest("LIST {0}", seq);
        return nconn.GetReplyReader ();
    }

    public NmapStreamReader BRAW(long seq, long start, long length)
    {
        nconn.WriteRequest("BRAW {0} {1} {2}", seq, start, length);
        return nconn.GetReplyReader ();
    }

    public NmapStreamReader BRAW(MessageInfo info)
    {
        return BRAW (info.seq, 0, info.bodyLength);
    }

    public string HEAD(int seq)
    {
        nconn.WriteRequest("HEAD {0}", seq);
        NmapResponse response = nconn.ReadLongReply ();
        return response.data;
    }

    public IList MIME(int seq)
    {
        nconn.WriteRequest("MIME {0}", seq);
        NmapResponse response = nconn.ReadMimeReply ();

        string[] lines = response.data.Split('\n');
        IList ret = new ArrayList(lines.Length);

        for(int i=0; i<lines.Length; i++)
        {
            if (lines[i].Equals(""))
            {
                continue;
            }

            string[] tmp = lines[i].Split(new char[]{' '}, 2);

            if (tmp[0].Equals("2003") || tmp[0].Equals("2004"))
            {
                continue;
            }

            ret.Add(GetMimeInfo(tmp[1]));
        }

        return ret;
    }
}
}
