using System;
using System.IO;
using System.Collections;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters.Binary;
using Hula.Nmap;
// using Hula.Ab;
using Hula.Cal;
using log4net;
using log4net.Config;

namespace Hula.Store {

public class FilteringStore {
    private static readonly ILog log = LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);

    private static Hula.Util.Filter nullFilter = new Hula.Util.Filter();
    private static Hashtable MimeFilters = new Hashtable();

    static FilteringStore() {
	/* xml -> iCal */
	MimeFilters["application/calendar+xml"] = new Hashtable();
	((Hashtable)MimeFilters["application/calendar+xml"])["text/calendar"] = new XmlFilterCal();

	/* iCal -> xml */
	MimeFilters["text/calendar"] = new Hashtable();
	((Hashtable)MimeFilters["text/calendar"])["application/calendar+xml"] = new CalFilterXml();

	/* these currently only work for vCard 2.1 */
#if false
	/* xml -> vCard */
	MimeFilters["application/vcard+xml"] = new Hashtable();
	((Hashtable)MimeFilters["application/vcard+xml"])["text/x-vcard"] = new XmlFilterCard();	

	/* vCard -> xml */
	MimeFilters["text/x-vcard"] = new Hashtable();
	((Hashtable)MimeFilters["text/x-vcard"])["application/vcard+xml"] = new CardFilterXml();
#endif
    }

    protected bool GetFilter(string contentType,
			     string accept,
			     out Hula.Util.Filter filter,
			     out string filteredContentType) {
	filter = nullFilter;
	filteredContentType = contentType;

	if (accept == null || accept == "" || accept == "*") {
	    return true;
	}

	Hashtable t = (Hashtable)MimeFilters[contentType];
	
	string[] split = accept.Split(new char[] {','});
	for (int i = split.Length - 1; i >= 0; i--) {
	    string s = split[i].Trim();
	    if (s == contentType) {
		return true;
	    } else if (t == null) {
		continue;
	    }

	    Hula.Util.Filter f = (Hula.Util.Filter)t[s];
	    if (f != null) {
		filteredContentType = s;
		filter = f;
		return true;
	    }
	}
	return false;
    }

    public string GetType (string contentType, string accept) {
	Hula.Util.Filter filter;
	return GetFilter (contentType, accept, out filter, out contentType) ? contentType : null;
    }

    protected TextReader Get (TextReader reader, string mimeType, string accept, out string retType) {
	Hula.Util.Filter filter;

	if (!GetFilter (mimeType, accept, out filter, out retType)) {
	    log.Warn ("Could not convert from " + mimeType + " to one of " + accept);
	    throw new System.Web.HttpException (405, "Method not allowed");
	}

	StringWriter writer = new StringWriter();
	filter.Run (reader, writer);
	return new StringReader (writer.ToString());
    }

    public TextReader Get (string guid, string accept, out string mimeType) {
	mimeType = ((IStore)this).GetProperty (guid, "hula.mime-type");
	using (TextReader reader = ((IStore)this).Get (guid)) {
	    return Get (reader, mimeType, accept, out mimeType);
	}
    }

    public void Get(string guid, string accept, out IDictionary properties, out TextReader reader, out string mimeType) {
	TextReader tmpReader;
	((IStore)this).Get (guid, out properties, out tmpReader);
	mimeType = (string)properties["hula.mime-type"];
	using (tmpReader) {
	    reader = Get (tmpReader, mimeType, accept, out mimeType);
	}
    }
}

}
