using System;
using System.Collections;

using log4net;

namespace Hula.Util
{

public class IndexingQueue
{
    private static readonly ILog log = LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);

    ArrayList NonInteractiveQueue = new ArrayList();
    ArrayList BulkQueue = new ArrayList();

    Hashtable Queued = new Hashtable();
    Hashtable Running = new Hashtable();

    public IndexingQueue()
    {
    }

    public void Put(ThreadInfo info)
    {
        lock(Queued)
        {
            ThreadInfo existingInfo = (ThreadInfo)Queued[info.UserData];

            if (existingInfo != null)
            {
                if (existingInfo.Priority < info.Priority)
                {
                    PromoteItem(info, existingInfo);
                }

                return;
            }

            InternalPut(info);
        }
    }

    public ThreadInfo Pop()
    {
        lock(Queued)
        {
            ArrayList[] queues = new ArrayList[]{NonInteractiveQueue, BulkQueue};

            foreach (ArrayList queue in queues)
            {
                for (int i=0; i<queue.Count; i++)
                {
                    ThreadInfo info = (ThreadInfo)queue[i];

                    if (!Running.ContainsKey(info.UserData))
                    {
                        Running[info.UserData] = info;
                        queue.RemoveAt(i);
                        Queued.Remove(info.UserData);
                        return info;
                    }
                }
            }

            return null;
        }
    }

    public void Finish(ThreadInfo info)
    {
        lock(Queued)
        {
            Running.Remove(info.UserData);
        }
    }

    public int Size()
    {
        lock(Queued)
        {
            return Queued.Count;
        }
    }

    private void InternalPut(ThreadInfo info)
    {
        GetQueueByInfo(info).Add(info);
        Queued[info.UserData] = info;
    }

    private ArrayList GetQueueByInfo(ThreadInfo info)
    {
        if (info.Priority == HulaThreadPriority.NonInteractive)
        {
            return NonInteractiveQueue;
        }
        else
        {
            return BulkQueue;
        }
    }

    private void PromoteItem(ThreadInfo info, ThreadInfo existingInfo)
    {
        ArrayList existingQueue = GetQueueByInfo(existingInfo);

        lock(Queued)
        {
            existingQueue.Remove(existingInfo);
            InternalPut(info);
        }
    }
}
}
