using System;
using System.Reflection;
using System.Threading;
using System.Collections;
using System.Net;
using System.Net.Sockets;
using System.IO;
using System.Web;
using System.Web.Hosting;
using System.Security.Cryptography;
using System.Security.Cryptography.X509Certificates;
using Mono.Security.Protocol.Tls;
using Mono.WebServer;
using SecurityProtocolType = Mono.Security.Protocol.Tls.SecurityProtocolType;
using X509Certificate = System.Security.Cryptography.X509Certificates.X509Certificate;
using Hula.Mdb;
using Hula.Nmap;
using Hula.Util;
using log4net;
using log4net.Config;

namespace Hula.Xsp {

public class HulaXsp {
    private static readonly ILog log = LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);

    int port;
    IPAddress addr;
    IWebSource webSource;
    ApplicationServer server;
    Hashtable applications = new Hashtable();

    static MemMgr memmgr = new MemMgr();

    public void ReadConfiguration() {
	// Defaults
	port = 8081;
	addr = IPAddress.Parse("0.0.0.0");

	string baseDir = System.AppDomain.CurrentDomain.BaseDirectory;
	applications["/"] = Path.Combine (baseDir, "htdocs");
        applications["/calcmd"] = Path.Combine (baseDir, "calcmd");
	applications["/dav"] = Path.Combine (baseDir, "dav");
	applications["/queue"] = Path.Combine (baseDir, "queue");
	applications["/search"] = Path.Combine (baseDir, "search");
	applications["/import"] = Path.Combine (baseDir, "import");

	
	string envApps = Environment.GetEnvironmentVariable("HULAWEB_EXTRA_APPS");
	if (envApps != null) {
	    string[] apps = envApps.Split(new char[] { ',' });
	    foreach (string app in apps) {
		Console.WriteLine("adding {0}", app);
		applications["/" + app] = Path.Combine(baseDir, app);
	    }
	}

	// FIXME: load settings from mdb
    }

    private void AddApplications() {
	foreach (string app in applications.Keys) {
	    server.AddApplication(null, port, app, (string)applications[app]);
	}
    }

    public int Run() {
	webSource = new XSPWebSource(addr, port);

	server = new ApplicationServer(webSource);
	AddApplications();
	
	if (server.Start(false) == false) {
	    return 2;
	}

	return 0;
    }

    public static int Main(string[] args) {
        HulaXsp xsp = new HulaXsp();
        xsp.ReadConfiguration();
        Logging.Configure();

        log.Info("started logging");

        return xsp.Run();
    }
}	
}
