using System;
using System.IO;
using System.Text;
using System.Collections;
using NUnit.Framework;
using Hula.ICal;
using Hula.TestData;

namespace Hula.Cal {

[TestFixture]
public class TestICal {
    public void CheckComponentTypes(ICalComponent obj, Hashtable expectedComponents) {
	ICalComponent component = obj.GetFirstComponent(ICalComponent.Kind.Any);
	while (component != null) {
	    string type = component.ComponentKind;
	    
	    if (expectedComponents.Contains(type)) {
		int newVal = (int)expectedComponents[type] - 1;

		if (newVal == 0) {
		    expectedComponents.Remove(type);
		} else {
		    expectedComponents[type] = newVal;
		}
	    } else {
		Assert.Fail("extra component of type: " + component.ComponentKind);
	    }
	    component = obj.GetNextComponent(ICalComponent.Kind.Any);	    
	}
	Assert.AreEqual(0, expectedComponents.Count);
    }

    [Test]
    public void TestSimpleICal() {
	ICalComponent obj = new ICalComponent(Data.GetReader("filter1.ics"));
	    
	Hashtable expectedComponents = new Hashtable();
	expectedComponents["VEVENT"] = 1;
	expectedComponents["VTIMEZONE"] = 1;

	CheckComponentTypes(obj, expectedComponents);
    }

    [Test]
    public void TestSimpleRecurrence() {
	ICalComponent obj = new ICalComponent(Data.GetReader("filter2.ics"));

	Hashtable expectedComponents = new Hashtable();
	expectedComponents["VEVENT"] = 6;

	CheckComponentTypes(obj, expectedComponents);
    }
}
}
