/*
 * <Novell-copyright>
 * Copyright (c) 2001 Novell, Inc. All Rights Reserved.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2 of the GNU General Public License
 * as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, contact Novell, Inc.
 * 
 * To contact Novell about this file by physical or electronic mail, you 
 * may find current contact information at www.novell.com.
 * </Novell-copyright>
 */
/*
    much of the code was copied verbatim from the file driver.  indeed the whole file structure
    was copied, and portions removed and edited as needed.

    function list left TODO

    MDBLDAPReadEx                   -- currently works, but not as advertised.  the function does the same thing as read
                                        because if it didn't then i'd have to hit the ldap server every time to get the
                                        next attribute.  Is that really desired over higher memory usage?
    MDBLDAPEnumerateObjectsEx           -- idem

    MDBLDAPWriteTyped               -- this function does not appear in the mdb or ddb docs
    MDBLDAPAddAddressRestriction            -- this function does not appear in the mdb or ddb docs
    MDBLDAPVerifyPasswordEx             -- this function does not appear in the mdb or ddb docs -- for now it just calls VerifyPassword()

    MDBLDAPClear                    -- may not function properly

    MDBLDAPEnumerateAttributesEx            -- commented out this code pending review of its necessity
    MDBLDAPRenameObject

    Ideas and issues:
        write a dn cache buffer that stores the mdb name and the ldap name of objects.
        this will allow the driver to be much, much faster in the long run as we won't
        have to re-search the ldap tree everytime, just some sort of hash or fifo.

        add connection pooling and possibly make it asynchronous

        finish cleaning up the code, removing un-needed portions left over from the file
        driver

        add the code to process arguments (setting the ldap server ip, port, if a cache
        is implemented, hash bucket size.

        add more than just simple plain text authentication

        error handling

        correctly checking if the context is valid prior to using it

*/
/* Product defines */
#define PRODUCT_NAME        "MDB OpenLDAP Driver"
#define PRODUCT_VERSION     "$Revision:   1.1  $"

#include <config.h>
#include <xpl.h>
#include <hulautil.h>

#include <errno.h>
#include <ctype.h>

#include <ldap.h>

#include "mdbldapp.h"

#ifdef WIN32
#include "free.h"   /* for Windows, contains ldap_mods_free() */
#endif              /* for Linux, ldap.h contains it          */

#include "ldif.h"

typedef struct _MDBldapClassStruct {
    unsigned char **names;
    unsigned char **naming;
    BOOL freeNaming;
    BOOL leafOnly;
    BOOL inheritContainment;

    struct _MDBldapClassStruct **subClasses;
    unsigned long subClassesUsed;
    struct _MDBldapClassStruct **containableClasses;
    unsigned long containableClassesUsed;
} MDBldapClassStruct;

struct  {
    BOOL unload;

    struct {
        XplSemaphore shutdown;
    } sem;

    struct {
        XplThreadID main;
        XplThreadID group;
    } id;

    /*  dirType = MDBLDAP_TYPE_EDIR, MDBLDAP_TYPE_OPENLDAP, MDBLDAP_TYPE_AD  */
    /*            or MDBLDAP_TYPE_NONE if not initialized.                   */
    unsigned long dirType; 
    unsigned char treeName[MDB_MAX_TREE_CHARS + 1];
    unsigned char rootDSE[MDB_MAX_OBJECT_CHARS + 1];  /*  do we need this?   */
    unsigned char *namingContexts[MDBLDAP_MAX_NAMING_CONTEXTS + 1];
    unsigned char subschemaSubentry[MDB_MAX_OBJECT_CHARS + 1]; /* for schema */
    unsigned char base64Chars[66];

    LDAP* ldap;
    BOOL bound;
    BOOL v3Support;

    /* just used for internal stuff.  never used against the db */
    MDBHandle mdbhandle;

    struct {
        BOOL enabled;
        BOOL console;

        FILE *fh;

        XplSemaphore sem;

        int bufferSize;

        unsigned char *buffer;
    } log;

    MDBldapClassStruct *schema;
} MDBldap;

typedef struct {
    char *mdbclassname;
    char *ldapclassname;
    char *naming;
} TranslationTableStruct;
TranslationTableStruct *translationtable=NULL;

#define ALLOC_AUTHENTICATED 30
MDBContextStruct **AuthenticatedUsers=NULL;
unsigned int Alloc_Authenticated=0;
unsigned int AuthenticatedCount=0;

char *DNAttributes[] = {
    "dn",
    NULL
};

char *ObjectClassAttributes[] = {
    "objectClass",
    NULL
};

char *ldapServer=NULL;
char *ldapUser=NULL;
char *ldapPassword=NULL;
char *ldapContext=NULL;
char *ldapSchemaPath=NULL;
FILE *outfile=NULL;
char *outfilepath=NULL;
char *cfgObject=NULL;
char serverDN[MDB_MAX_OBJECT_CHARS+1];
int debuglevel=0;
int ldapPort=0;

/* Internal Functions */
BOOL GetLDAPServerInfo(LDAP *ldap);
MDBContextStruct *MDBLDAPFindHandle(MDBContextStruct *in);
MDBContextStruct *MDBLDAPAddHandle(LDAP *newLDAP);
BOOL MDBLDAPRemoveHandle(MDBContextStruct *in);
char *MDBLDAPFindNaming(const char *Class);
int MDBLDAPldap2mdb(const char *Input, char *Output, const char *BaseDN);
void MDBLDAPFindObjectR(const char *Object, const char *BaseDN, char **result, MDBValueStruct *V);
char *MDBLDAPFindObject(const char *Object, const char *BaseDN, MDBValueStruct *V);
long MDBLDAP_Read(const unsigned char *Object, const unsigned char *Attribute, MDBValueStruct *V, BOOL isDN);
BOOL MDBLDAP_Write(const unsigned char *Object, const unsigned char *Attribute, MDBValueStruct *V, int Type, BOOL isDN);
BOOL MDBLDAPParseArguments(const unsigned char *Arguments);
void RemoveSpaces(char *val);
void LoadTranslationTable(unsigned char *path);
void FreeTranslationTable(void);
unsigned char *FindLDAPClass(const unsigned char *mdbclass);
unsigned char *FindMDBClass(const unsigned char *ldapclass);

BOOL MDBParseInit(const unsigned char *Arguments);
BOOL MDBParseGetNext(unsigned char **Name, unsigned char **Value);
BOOL MDBParseEnd(void);

/* MDB Functions */
BOOL MDBLDAPGetServerInfo(unsigned char *HostDN, unsigned char *HostTree, MDBValueStruct *V);
MDBHandle MDBLDAPAuthenticate(const unsigned char *Object, const unsigned char *Password, const unsigned char *Arguments);
BOOL MDBLDAPRelease(MDBHandle Handle);
MDBValueStruct* MDBLDAPCreateValueStruct(MDBHandle Handle, const unsigned char *Context);
BOOL MDBLDAPDestroyValueStruct(MDBValueStruct *V);
MDBValueStruct* MDBLDAPShareContext(MDBValueStruct *V);
BOOL MDBLDAPSetValueStructContext(const unsigned char *Context, MDBValueStruct *V);
MDBEnumStruct* MDBLDAPCreateEnumStruct(MDBValueStruct *V);
BOOL MDBLDAPDestroyEnumStruct(MDBEnumStruct *E, MDBValueStruct *V);
BOOL MDBLDAPAddValue(const unsigned char *Value, MDBValueStruct *V);
BOOL MDBLDAPFreeValue(unsigned long Index, MDBValueStruct *V);
BOOL MDBLDAPFreeValues(MDBValueStruct *V);
long MDBLDAPRead(const unsigned char *Object, const unsigned char *Attribute, MDBValueStruct *V);
const unsigned char* MDBLDAPReadEx(const unsigned char *Object, const unsigned char *Attribute, MDBEnumStruct *E, MDBValueStruct *V);
long MDBLDAPReadDN(const unsigned char *Object, const unsigned char *Attribute, MDBValueStruct *V);
BOOL MDBLDAPWriteTyped(const unsigned char *Object, const unsigned char *Attribute, const int AttrType, MDBValueStruct *V);
BOOL MDBLDAPWrite(const unsigned char *Object, const unsigned char *Attribute, MDBValueStruct *V);
BOOL MDBLDAPWriteDN(const unsigned char *Object, const unsigned char *Attribute, MDBValueStruct *V);
BOOL MDBLDAPClear(const unsigned char *Object, const unsigned char *Attribute, MDBValueStruct *V);
BOOL MDBLDAPAdd(const unsigned char *Object, const unsigned char *Attribute, const unsigned char *Value, MDBValueStruct *V);
BOOL MDBLDAPAddDN(const unsigned char *Object, const unsigned char *Attribute, const unsigned char *Value, MDBValueStruct *V);
BOOL MDBLDAPRemove(const unsigned char *Object, const unsigned char *Attribute, const unsigned char *Value, MDBValueStruct *V);
BOOL MDBLDAPRemoveDN(const unsigned char *Object, const unsigned char *Attribute, const unsigned char *Value, MDBValueStruct *V);
BOOL MDBLDAPIsObject(const unsigned char *Object, MDBValueStruct *V);
BOOL MDBLDAPGetObjectDetails(const unsigned char *Object, unsigned char *Type, unsigned char *RDN, unsigned char *DN, MDBValueStruct *V);
BOOL MDBLDAPVerifyPassword(const unsigned char *Object, const unsigned char *Password, MDBValueStruct *V);
BOOL MDBLDAPVerifyPasswordEx(const unsigned char *Object, const unsigned char *Password, MDBValueStruct *V);
BOOL MDBLDAPChangePassword(const unsigned char *Object, const unsigned char *OldPassword, const unsigned char *NewPassword, MDBValueStruct *V);
BOOL MDBLDAPChangePasswordEx(const unsigned char *Object, const unsigned char *OldPassword, const unsigned char *NewPassword, MDBValueStruct *V);
long MDBLDAPEnumerateObjects(const unsigned char *Container, const unsigned char *Type, const unsigned char *Pattern, MDBValueStruct *V);
const unsigned char* MDBLDAPEnumerateObjectsEx(const unsigned char *Container, const unsigned char *Type, const unsigned char *Pattern, unsigned long Flags, MDBEnumStruct *E, MDBValueStruct *V);
BOOL MDBLDAPAddAddressRestriction(const unsigned char *Object, const unsigned char *Server, MDBValueStruct *V);
const unsigned char* MDBLDAPEnumerateAttributesEx(const unsigned char *Object, MDBEnumStruct *E, MDBValueStruct *V);
BOOL MDBLDAPCreateObject(const unsigned char *Object, const unsigned char *Class, MDBValueStruct *Attribute, MDBValueStruct *Data, MDBValueStruct *V);
BOOL MDBLDAPDeleteObject(const unsigned char *Object, BOOL Recursive, MDBValueStruct *V);
BOOL MDBLDAPRenameObject(const unsigned char *ObjectOld, const unsigned char *ObjectNew, MDBValueStruct *V);
BOOL MDBLDAPDefineAttribute(const unsigned char *Attribute, const unsigned char *ASN1, unsigned long Type, BOOL SingleValue, BOOL ImmediateSync, BOOL Public, MDBValueStruct *V);
BOOL MDBLDAPUndefineAttribute(const unsigned char *Attribute, MDBValueStruct *V);
BOOL MDBLDAPDefineClass(const unsigned char *Class, const unsigned char *ASN1, BOOL Container, MDBValueStruct *Superclass, MDBValueStruct *Containment, MDBValueStruct *Naming, MDBValueStruct *Mandatory, MDBValueStruct *Optional, MDBValueStruct *V);
BOOL MDBLDAPAddAttribute(const unsigned char *Attribute, const unsigned char *Class, MDBValueStruct *V);
BOOL MDBLDAPUndefineClass(const unsigned char *Class, MDBValueStruct *V);
BOOL MDBLDAPListContainableClasses(const unsigned char *Container, MDBValueStruct *V);
const unsigned char* MDBLDAPListContainableClassesEx(const unsigned char *Container, MDBEnumStruct *E, MDBValueStruct *V);
BOOL MDBLDAPGrantObjectRights(const unsigned char *Object, const unsigned char *TrusteeDN, BOOL Read, BOOL Write, BOOL Delete, BOOL Rename, BOOL Admin, MDBValueStruct *V);
BOOL MDBLDAPGrantAttributeRights(const unsigned char *Object, const unsigned char *Attribute, const unsigned char *TrusteeDN, BOOL Read, BOOL Write, BOOL Admin, MDBValueStruct *V);
EXPORT BOOL MDBLDAPInit(MDBDriverInitStruct *Init);
EXPORT void MDBLDAPShutdown(void);

unsigned char *Args;
unsigned char *ArgsEnd;
unsigned char *ptr=NULL;
unsigned char *end=NULL;

BOOL
MDBParseInit(const unsigned char *Arguments)
{
    Args = (unsigned char *)strdup((const char *)Arguments);
    ArgsEnd = Args+strlen((const char *)Args);
    return TRUE;
}

BOOL
MDBParseGetNext(unsigned char **Name, unsigned char **Value)
{
    unsigned char *tmpName=NULL;

    if (!ptr || !end) {
        ptr = end = Args;
    }

    if (ptr >= ArgsEnd) {
        return FALSE;
    }

    /* move end till we get to the = sign and store the result in Name */
    end = (unsigned char *)strchr((const char *)ptr, '=');
    if (end) {
        *end = '\0';
        tmpName = ptr;

        ptr = end + 2; /* eat the null and the quote */
        end = ptr;
        end = (unsigned char *)strchr((const char *)ptr, '"');
        if (end) {
            *end = '\0';

            *Name = tmpName;
            *Value = ptr;

            ptr = end + 2; /* eat the null and the comma */
        }
    }
    return TRUE;
}

BOOL
MDBParseEnd(void)
{
    if (Args) {
        free(Args);
    }
    return TRUE;
}

/* Internal Functions */
void
RemoveSpaces(char *val)
{
    char *at, *next;

    at = val;
    next = at;
    while(*next) {
        while(*next && (*next == ' ' || *next == ':')) {
            next++;
        }
        *at++ = *next++;
    }
    *at = '\0';
    return;
}

/*
    GetLDAPServerInfo() retrieves various attribute values from the root DSE
    of the LDAP server, so we can initialize ourselves.  The root DSE contains
    LDAP-server-specific information.  Get all the information we need now,
    so we don't have to hit the wire multiple times.
    If we've already done this before, don't do it again - return TRUE.
*/
BOOL
GetLDAPServerInfo(LDAP *ldap)
{
    int i, err, refs = -1;
    char *a;
    char **vals;
    LDAPMessage *pres, *res = NULL;
    BerElement *ber = NULL;
    char *attrs[] = {
        "supportedLDAPVersion",
        "namingContexts",
        "directoryTreeName",
        "vendorName",
        "subschemaSubentry",  /*  has the dn of the schema attribute.        */
        /*  In the future, we may want/need to check these attributes        */
        /* "vendorVersion",                                                  */
        /* "supportedExtension",                                             */
        /* "supportedControl",                                               */
        NULL
    };

    if (MDBldap.dirType == MDBLDAP_TYPE_NONE) {
        /*  To get the root DSE, we must turn off referral handling.        */
        /*  But we'll restore it after we're done.                          */
        ldap_get_option(ldap, LDAP_OPT_REFERRALS, (void *)(&refs));
        ldap_set_option(ldap, LDAP_OPT_REFERRALS, LDAP_OPT_OFF);

        err = ldap_search_ext_s(ldap, "", LDAP_SCOPE_BASE, "(objectClass=*)",
                            attrs, 0, NULL, NULL, NULL, 0, &pres);
        if (err == LDAP_SUCCESS) {
            /*  There should only be one entry that matches.                 */
            if ((res = ldap_first_entry(ldap, pres))) {
                for (a = ldap_first_attribute(ldap, res, &ber) ;
                         a != NULL ;
                         a = ldap_next_attribute(ldap, res, ber)) {
                    if ((vals = ldap_get_values(ldap, res, a)) != NULL) {
                        for (i = 0 ; vals[i] != NULL; i++ ) {
                            if (strcmp("supportedLDAPVersion", a) == 0) {
                                /*  eDirectory 8.7.3 supports v2 & v3.       */
                                if (strcmp("3", vals[i]) == 0) {
                                    MDBldap.v3Support = TRUE;
                                    continue;
                                }
                            }
                            if (strcmp("directoryTreeName", a) == 0) {
                                /*  eDirectory always has this w/ a value.   */
                                /*  OpenLDAP never has this attribute.       */
                                /*  Does AD?                                 */
                                MDBldap.dirType = MDBLDAP_TYPE_EDIR;
                                strcpy(MDBldap.treeName, vals[i]);
                                /*  We assume there's just one, if there's   */
                                /*  more, we'll just use the last one.       */
                                continue;
                            }
                            if (strcmp("vendorName", a) == 0) {
                                /*  eDirectory always has this w/ a value.   */
                                if (strcmp(vals[i], "Novell, Inc.") == 0) {
                                    MDBldap.dirType = MDBLDAP_TYPE_EDIR;
                                } else {
                                    if (strcmp(vals[i], "OpenLDAP") == 0) {
                                        MDBldap.dirType = MDBLDAP_TYPE_OPENLDAP;
                                    }
                                }
                                continue;
                            }
                            if (strcmp("namingContexts", a) == 0) {
                                /*  Always typed.      */
                                /*  eDirectory has this attr w/ no value.    */
                                /*  (at least my eDirs have had none).       */
                                /*  OpenLDAP always has this w/ a value.     */
                                /*  It may have more than 1 value.           */
                                if (vals[i] != NULL && vals[i][0] != '\0') {
                                   MDBldap.dirType = MDBLDAP_TYPE_OPENLDAP;
                                   if (i < MDBLDAP_MAX_NAMING_CONTEXTS) {
                                      MDBldap.namingContexts[i]=strdup(vals[i]);
                                   }
                                }
                                continue;
                            }
                            if (strcmp("subschemaSubentry", a) == 0) {
                                /*  Always typed.                            */
                                /*  Always has a value?                      */
                                if (vals[i] != NULL && vals[i][0] != '\0') {
                                    strcpy(MDBldap.subschemaSubentry, vals[i]);
                                }
                                continue;
                            }
                            /* DEBUG: printf ("%s: %s\n", a, vals[i]);       */
                        }
                        ldap_value_free(vals);
                    }
                    ldap_memfree(a);
                }

                /*  Free memory allocated by ldap_first_attribute().         */
                if (ber != NULL) {
                    ber_free(ber, 0);
                }
            }
        }   /*  End  getting root DSE  */

        /*  Restore the LDAP Session setting for referrals.                  */
        if (refs == 0) {
            ldap_set_option(ldap, LDAP_OPT_REFERRALS, LDAP_OPT_OFF);
        } else {
            ldap_set_option(ldap, LDAP_OPT_REFERRALS, LDAP_OPT_ON);
        }

    } else {
        return (TRUE);                        /*  we've already initialized  */
    }

    /*  If the root DSE was not found, the server is not V3-compliant.       */
    if (!MDBldap.v3Support || MDBldap.dirType == MDBLDAP_TYPE_NONE) {
        /*  We'll free any namingContexts at Shutdown() time.                */
        return (FALSE);
    }

    if (MDBldap.dirType == MDBLDAP_TYPE_EDIR && MDBldap.treeName[0] == '\0') {
        strcpy(MDBldap.treeName, "[Root]");     /*  default for eDir         */
    }

    return (TRUE);

}  /*  End  GetLDAPServerInfo()  */


/*
    MDBLDAPFindHandle searches the list of authenticated users for the give struct.
    returns NULL if it is not found and the object if it is
*/
MDBContextStruct *
MDBLDAPFindHandle(MDBContextStruct *in)
{
    MDBContextStruct *result=NULL;
    int x=0;

    while ((result = AuthenticatedUsers[x++])) {
        if (in == result) {
            return result;
        }
    }
    return NULL;
}

/*
    MDBLDAPAddHandle adds a newly authenticated user to the list of authenticated users
    Returns a new MDBContextStruct if successful, NULL if not.  This is what is returned
    to the client app for it to pass back to us for use in almost all the functions
*/
MDBContextStruct *
MDBLDAPAddHandle(LDAP *newLDAP)
{
    MDBContextStruct *in;

    if (AuthenticatedCount+1 > Alloc_Authenticated) {
        /* re-alloc to add space for the new user */
        Alloc_Authenticated += ALLOC_AUTHENTICATED;
        AuthenticatedUsers = realloc(AuthenticatedUsers, (Alloc_Authenticated)*sizeof(MDBHandle));
    }

    in = malloc(sizeof(MDBContextStruct));
    in->bindhandle = newLDAP;
    XplOpenLocalSemaphore(in->sem, 1);

    if (AuthenticatedUsers) {
        AuthenticatedUsers[AuthenticatedCount-1] = in;
        AuthenticatedUsers[AuthenticatedCount++] =  NULL;
        return in;
    }
    return NULL;
}

/*
    MDBLDAPRemoveHandle removes an authenticated user from the list and free's its memory
*/
BOOL
MDBLDAPRemoveHandle(MDBContextStruct *in)
{
    /* first find the handle */
    MDBContextStruct *result=NULL;
    int x=0;

    while ((result = AuthenticatedUsers[x])) {
        if (in == result) {
            XplWaitOnLocalSemaphore(result->sem);
            ldap_unbind_s(result->bindhandle);
            XplCloseLocalSemaphore(result->sem);
            free(result);
            /* move everything down */
            memmove(&AuthenticatedUsers[x], &AuthenticatedUsers[x+1], sizeof(MDBContextStruct *)*(AuthenticatedCount-1-x));
            AuthenticatedCount--;

            if ((AuthenticatedCount+30) < Alloc_Authenticated) {
                Alloc_Authenticated -= ALLOC_AUTHENTICATED;
                AuthenticatedUsers = realloc(AuthenticatedUsers, (Alloc_Authenticated)*sizeof(MDBContextStruct *));
            }
            return TRUE;
        }
        x++;
    }
    return FALSE;
}

/*
    MDBLDAPFindNaming is used in createobject to determine the prefix for the ldap dn

    It will first try to find the naming prefix in the schema table (as read from LDAP, and
    use a value found there if possible.  If not then fall back to the translation table.
*/
char *
MDBLDAPFindNaming(const char *Class)
{
    unsigned long cur;

    cur = 0;
    while (MDBldap.schema[cur].names) {
        unsigned long n = 0;

        while (MDBldap.schema[cur].names[n]) {
            if (MDBldap.schema[cur].naming && MDBldap.schema[cur].naming[0] &&
                XplStrCaseCmp(MDBldap.schema[cur].names[n], Class) == 0)
            {
                return(MDBldap.schema[cur].naming[0]);
            }
            n++;
        }
        cur++;
    }

    cur = 0;
    while (translationtable[cur].mdbclassname != NULL) {
        if (XplStrCaseCmp(translationtable[cur].mdbclassname, Class) == 0) {
            if (!translationtable[cur].naming || strlen(translationtable[cur].naming) == 0) {
                return "CN";
            } else {
                return translationtable[cur].naming;
            }
        }
        cur++;
    }

    /* Default to CN.  Its likely to work better than NULL */
    return("CN");
}

/*
    converts an ldap name to an MDB name and returns its length
    This function must be called prior to returning any DN to
      mdb proper as mdb -- and all clients that use use mdb --
      are expecting the following format \\o\ou\ou\cn

    Input is the buffer containing the name to be converted
    Output is a pointer to the buffer to receive the output -- no length checking is done
    BaseDN, if passed, is the DN to try to make things relative to
*/
int
MDBLDAPldap2mdb(const char *Input, char *Output, const char *BaseDN)
{
    int len=0;
    char *cur;
    char **inputar, **basednar;
    int inputarp, basednarp;

    /* make sure to only do this on DN attributes */
    if ((cur = strchr(Input, '='))) {
        inputar=ldap_explode_dn(Input, 1);
        /* find the end item in the array */
        for (inputarp=0;inputar[inputarp+1]!=NULL;inputarp++) {
            /* do nothing */
        }
        if (BaseDN) {
            basednar=ldap_explode_dn(BaseDN, 1);

            /* find th eend item in the array */
            for (basednarp=0;basednar[basednarp+1]!=NULL;basednarp++) {
                /* do nothing */
            }

            /* work our way back through the array to get to the first non-equivalent entry */
            while (XplStrCaseCmp(inputar[inputarp], basednar[basednarp]) == 0) {
                inputarp--;
                basednarp--;
                if (inputarp < 0 || basednarp < 0) {
                    break;
                }
            }

            /* 
             * when we get here we are on the first non-equivalent entries
             *   so that means that in the input array we are on the first
             *   relative context.  if the count is more than one we need
             *   to use slashes to separate the items.  otherwise just return
             *   then item
            */
            if (inputarp == 0) {
                if (inputar[inputarp] != NULL) {
                    sprintf(Output, "%s", inputar[0]);
                }
            } else {
                /* now step through inputar backwards adding to output as we go */
                while (inputarp >= 0) {
                    len+=sprintf(Output+len, "%s\\", inputar[inputarp--]);
                }
                Output[len] = '\0';
            }
            ldap_value_free(basednar);
        } else {
            /* now step through inputar backwards adding to output as we go */
            while (inputarp >= 0) {
                len+=sprintf(Output+len, "\\%s", inputar[inputarp--]);
            }
        }
        ldap_value_free(inputar);
    } else {
        len=sprintf(Output, "%s", Input);
    }
    
    return len;
}

/*
    This recursive function parses the ENTIRE tree searching for the desired object

    this is TERRIBLY inefficient, but this is the only way I can
     think of to do this:
       search starting from root for the individual pieces
          upon finding them, find it's dn.

    Object is a straight name e.g. Internet Services
    BaseDN is an ldap style named object e.g. o=Internet Services
*/
void
MDBLDAPFindObjectR(const char *Object, const char *BaseDN, char **result, MDBValueStruct *V) {
    char *object = NULL, *objectdn = NULL;
    char **rdn, *bptr = NULL, *eptr = NULL;
    char *dn;
    int err;
    LDAPMessage *pres = NULL;
    LDAPMessage *pres2 = NULL;
    BOOL found = FALSE, namingContext = FALSE;

    if (!Object) {
        return;
    }

    /* If the object starts with a slash and the treename, then cut that portion off */
    if (Object[0] == '\\' && strncmp(MDBldap.treeName, Object + 1, strlen(MDBldap.treeName)) == 0) {
        if (Object[strlen(MDBldap.treeName) + 1] != '\0') {
            object = strdup(Object + strlen(MDBldap.treeName) + 1);
        } else {
            /*  Take care of the case where we are searching for             */
            /*  the object that happens to be the name of the tree.          */
            if (MDBldap.rootDSE != NULL) {
                object = strdup(MDBldap.treeName);
                namingContext = TRUE;
            } else {
                object = strdup("\\");
            }
        }
    } else {
        object = strdup(Object);
    }

    /* first find the section of Object in question */
    /* strip off any leading \ chars */
    bptr = object;
    while (*bptr == '\\') {
        bptr++;
    }

    eptr = strchr(bptr, '\\');
    if (eptr) {
        *eptr = '\0';
    }

    /* do an ldap search in BaseDN for all objects */
    XplWaitOnLocalSemaphore(V->context->sem);
    if (namingContext) {    /*  We need to do this on OpenLDAP.             */
        err = ldap_search_s(V->context->bindhandle, BaseDN,
                LDAP_SCOPE_BASE, "(objectClass=*)", DNAttributes, 0, &pres);
    } else {                /*  This works for eDirectory.                  */
        err = ldap_search_s(V->context->bindhandle, BaseDN,
                LDAP_SCOPE_ONELEVEL, "(objectClass=*)", DNAttributes, 0, &pres);
    }
    if (err==LDAP_SUCCESS) {
        pres2 = ldap_first_entry(V->context->bindhandle, pres);
        while (pres2) {
            objectdn = ldap_get_dn(V->context->bindhandle, pres2);
            rdn = ldap_explode_dn(objectdn, 0);
            dn = strchr(rdn[0], '=') + 1;
            if (strcmp(bptr, dn) == 0) {
                found = TRUE;
                *result = realloc(*result, (strlen(objectdn) + 1) * sizeof(char));
                sprintf(*result, "%s", objectdn);
                ldap_value_free(rdn);
                ldap_memfree(objectdn);
                break;
            }
            ldap_value_free(rdn);
            ldap_memfree(objectdn);
            pres2 = ldap_next_entry(V->context->bindhandle, pres2);
        }
    } else {
        printf("ldap_search_s() failed with %s\n", ldap_err2string(err));
    }

    if (!found) {
        **result = '\0';
    }

    if (pres) {
        ldap_msgfree(pres);
    }

    if (eptr) {
        *eptr = '\\';
    }
    XplSignalLocalSemaphore(V->context->sem);
    MDBLDAPFindObjectR(eptr, *result, result, V);
    free(object);
}

char *
MDBLDAPFindObject(const char *Object, const char *BaseDN, MDBValueStruct *V) {
    char *result;
    
    if (!Object || !V || !MDB_CONTEXT_IS_VALID(V)) {
        return NULL;
    }
        
    result = malloc(strlen(Object)+1);
    *result = '\0';
    /* I only want to use the BaseDN if the first two chars of Object are not \\ */
    if (Object[0] != '\\' && Object[1] != '\\') {
        MDBLDAPFindObjectR(Object, BaseDN, &result, V);
    } else {
        MDBLDAPFindObjectR(Object, MDBldap.rootDSE, &result, V);
    }
    return result;
}

/*
    I use this function to actually read from an object because in the LDAP
    world Read and ReadDN are nearly identical in code.  This adds an additional
    function call, but it saves code duplication.

    Read, if used to read an attribute that contains a DN, will only return an absolute DN
      if the DN has no parents in common with Object

    ReadDN, if used to read an attribute that contains a DN, will ALWAYS return an absolute DN
*/
long
MDBLDAP_Read(const unsigned char *Object, const unsigned char *Attribute, MDBValueStruct *V, BOOL isDN)
{
    char *filter[2];
    int err;
    LDAPMessage *pres = NULL;
    LDAPMessage *pres2 = NULL;
    char **values=NULL;
    char *currentValue;
    char *obj;
    long result = 0;
    char tmpValue[MDB_MAX_OBJECT_CHARS + 1];
    char *attribute = NULL;

    if (!V || !MDB_CONTEXT_IS_VALID(V)) {
        return result;
    }

    obj = MDBLDAPFindObject(Object, V->BaseDN, V);
    if (obj[0] != '\0' || Object[0] == '\0') {
        attribute = strdup(Attribute);
        RemoveSpaces(attribute);
        filter[0] = attribute;
        filter[1] = NULL;

        XplWaitOnLocalSemaphore(V->context->sem);
        err = ldap_search_s(V->context->bindhandle, obj, LDAP_SCOPE_BASE, "(objectClass=*)", filter, 0, &pres);
        if (err==LDAP_SUCCESS) {
            pres2 = ldap_first_entry(V->context->bindhandle, pres);
            if (pres2) {
                values = ldap_get_values(V->context->bindhandle, pres2, attribute);
                if (values) {
                    while ((currentValue = values[result]) != NULL) {
                        if (isDN) {
                            MDBLDAPldap2mdb(currentValue, tmpValue, NULL);
                            MDBLDAPAddValue(tmpValue, V);
                        } else {
                            MDBLDAPAddValue(currentValue, V);
                        }
                        /* printf("%s\n", tmpValue); */
                        result++;
                    }
                }
            }
            if (values) {
                ldap_value_free(values);
            }
        }
        if (pres) {
            ldap_msgfree(pres);
        }
        XplSignalLocalSemaphore(V->context->sem);
        free(attribute);
    }

    free(obj);
    return result;
}

BOOL
MDBLDAP_Write(const unsigned char *Object, const unsigned char *Attribute, MDBValueStruct *V, int mod_op, BOOL isDN)
{
    char *dn;
    LDAPMod **attrs;
    unsigned long x;
    int err;
    BOOL result=FALSE;
    char *tmpValue;

    dn = MDBLDAPFindObject(Object, V->BaseDN, V);
    
    if (strlen(dn) > 0) {
        /* set the ldapmod up */
        attrs = malloc(2*sizeof(LDAPMod *));
        attrs[0] = malloc(sizeof(LDAPMod));
        attrs[0]->mod_op = mod_op | LDAP_MOD_BVALUES;
        attrs[0]->mod_type = strdup(Attribute);
        RemoveSpaces(attrs[0]->mod_type);
        attrs[0]->mod_bvalues = malloc((V->Used+1) * sizeof(struct berval *));
        for(x=0;x<V->Used;x++) {
            if (isDN) {
                tmpValue = MDBLDAPFindObject(V->Value[x], V->BaseDN, V);
            } else {
                tmpValue = strdup(V->Value[x]);
            }
            attrs[0]->mod_bvalues[x] = malloc(sizeof(struct berval));
            attrs[0]->mod_bvalues[x]->bv_len = strlen(tmpValue);
            attrs[0]->mod_bvalues[x]->bv_val = tmpValue;
        }
        attrs[0]->mod_bvalues[V->Used] = NULL;
        attrs[1] = NULL;

        XplWaitOnLocalSemaphore(V->context->sem);
        err = ldap_modify_s(V->context->bindhandle, dn, attrs);
        XplSignalLocalSemaphore(V->context->sem);
        if (err == LDAP_SUCCESS) {
            result = TRUE;
        }

        for(x=0;x<V->Used;x++) {
            free(attrs[0]->mod_bvalues[x]->bv_val);
            free(attrs[0]->mod_bvalues[x]);
        }
        free(attrs[0]->mod_bvalues);
        free(attrs[0]);
        free(attrs);
    }
    free(dn);
    return result;
}

BOOL
MDBLDAPParseArguments(const unsigned char *Arguments)
{
    unsigned char *Name;
    unsigned char *Value;

    MDBParseInit(Arguments);
    while (MDBParseGetNext(&Name, &Value)) {
        if (XplStrNCaseCmp(Name, "server", 6) == 0) {
            ldapServer = strdup(Value);
        } else if (XplStrNCaseCmp(Name, "user", 4) == 0) {
            ldapUser = strdup(Value);
        } else if (XplStrNCaseCmp(Name, "password", 8) == 0) {
            ldapPassword = strdup(Value);
        } else if (XplStrNCaseCmp(Name, "context", 7) == 0) {
            ldapContext = strdup(Value);
        } else if (XplStrNCaseCmp(Name, "debug", 5) == 0) {
            debuglevel = atoi(Value);
        } else if (XplStrNCaseCmp(Name, "translation", 11) == 0) {
            ldapSchemaPath = strdup(Value);
        } else if (XplStrNCaseCmp(Name, "schemaout", 9) == 0) {
            outfilepath = strdup(Value);
        } else if (XplStrNCaseCmp(Name, "port", 4) == 0) {
            ldapPort = atoi(Value);
        } else if (XplStrNCaseCmp(Name, "cfgobject", 9) == 0) {
            cfgObject = strdup(Value);
        }
    }

    if (ldapPort==0) {
        ldapPort = LDAP_PORT;
    }

    MDBParseEnd();

    /* the ldapSchemaPath needs to be semi optional otherwise you can never generate one
       either ldapContext or cfgObject must be set */
    return(ldapServer && ldapUser && ldapPassword && ldapPort && (ldapContext || cfgObject));
}

/*
typedef struct {
            char *mdbclassname;
                char *ldapclassname;
                    char *naming;
} TranslationTableStruct;
TranslationTableStruct *translationtable=NULL;
*/

#ifndef WIN32
void
LoadTranslationTable(unsigned char *path)
{
    LDIFFP *p;
    int lno=0;
    char *buffer, *bufchange;
    char *name, *value;
    char *line;
    int length=8000;
    int result;
    ber_len_t vlenp;
    int allocated=0;

    p = ldif_open(path, "r");
    if (p) {
        buffer = malloc(8000); /* this is way too big for what we need, but better safe than sorry, and it only happens once */
        bufchange = buffer;
        while ((result = ldif_read_record(p, &lno, &buffer, &length)) == 1) {
            while ((line = ldif_getline(&bufchange)) != NULL) {
                translationtable = realloc(translationtable, sizeof(TranslationTableStruct)*(allocated+1));
            
                ldif_parse_line(line, &name, &value, &vlenp);
                if (XplStrCaseCmp("mdbclass", name) == 0) {
                    translationtable[allocated].mdbclassname = value;
                } else if (XplStrCaseCmp("ldapclass", name) == 0) {
                    translationtable[allocated].ldapclassname = value;
                } else if (XplStrCaseCmp("naming", name) == 0) {
                    translationtable[allocated].naming = value;
                }
                /* make sure to free name because i don't need to keep it around */
                /*   i do, however, need the value so don't free that till later */
                //free(name);
            }
            allocated++;
            length=8000;
            bufchange = buffer;
        }
        free(buffer);
        ldif_close(p);
    }
    /* allocate one more pointer to terminate */
    translationtable = realloc(translationtable, (sizeof *translationtable)*(allocated+1));
    translationtable[allocated].mdbclassname = NULL;
    translationtable[allocated].ldapclassname = NULL;
    translationtable[allocated].naming = NULL;
    return;
}
#endif

void
FreeTranslationTable(void)
{
    int cur=0;
    
    while(translationtable[cur].mdbclassname != NULL) {
        free(translationtable[cur].mdbclassname);
        free(translationtable[cur].ldapclassname);
        free(translationtable[cur].naming);
        cur++;
    }
    free(translationtable);
}

unsigned char *
FindLDAPClass(const unsigned char *mdbclass)
{
    /* search the MDBClass for the passed in value, then take its ldap class */
    unsigned long cur=0;
    while(translationtable[cur].mdbclassname != NULL) {
        if (XplStrCaseCmp(translationtable[cur].mdbclassname, mdbclass) == 0) {
            return translationtable[cur].ldapclassname;
        }
        cur++;
    }
    return NULL;
}

unsigned char *
FindMDBClass(const unsigned char *ldapclass)
{
    /* search the LDAPClass for the passed in value, then take its mdb class */
    unsigned long cur=0;
    while (translationtable[cur].ldapclassname != NULL) {
        if (XplStrCaseCmp(translationtable[cur].ldapclassname, ldapclass) == 0) {
            return translationtable[cur].mdbclassname;
        }
        cur++;
    }
    return NULL;
}

/* MDB Functions */
/* GetServerInfo is called prior to almost every other call.  why, i'm not sure */  
BOOL
MDBLDAPGetServerInfo(unsigned char *HostDN, unsigned char *HostTree, MDBValueStruct *V)
{
    if (HostDN) {
        if (cfgObject) {
            if (strchr(cfgObject, '=')) {
                /* if the cfg object is ldap format convert it */
                MDBLDAPldap2mdb(cfgObject, HostDN, NULL);
            } else {
                /* if the cfgObject is specified and it is mdb format, just return it */
                strcpy(HostDN, cfgObject);
            }
        } else {
            /* we don't have a cfgObject, so we need to fall back to the only thing left to us */
            MDBLDAPldap2mdb(serverDN, HostDN, NULL);
        }
    }

    if (HostTree) {
        if (MDBldap.dirType == MDBLDAP_TYPE_EDIR) {
            strcpy(HostTree, MDBldap.treeName);
        } else {
            strcpy(HostTree, MDBldap.rootDSE);
        }
    }
    return TRUE;
}

MDBHandle
MDBLDAPAuthenticate(const unsigned char *Object, const unsigned char *Password, const unsigned char *Arguments)
{
    int err, version = LDAP_VERSION3;
    /* DEBUG: int debugLevel = -1;  (ALL) */
    LDAP *ldap;
    unsigned char *ldapuname;
    MDBContextStruct *context;

    /*  We get protocol errors on OpenLDAP if version 3 is not set.       */
    /*  Setting this before the ldap_init() with a NULL handle, causes    */
    /*  the option to be set for all sessions we create afterwards.       */
    ldap_set_option(NULL, LDAP_OPT_PROTOCOL_VERSION, &version);
    /* DEBUG: ldap_set_option(ldap, LDAP_OPT_DEBUG_LEVEL, &debugLevel);   */

    /*  ldap_open() will be deprecated in a later release, so we're using
        ldap_init() which is now recommended.  The difference is that a
        connection is not opened with ldap_init(), but it will be opened
        when the first operation is attempted.
    */
    if ((ldap = ldap_init(ldapServer, ldapPort)) == (LDAP *)0) {
        return (MDBHandle)0;
    }

    if (Object == NULL) {
        context = MDBLDAPAddHandle(ldap);
        return (MDBHandle)(context);
    }

    /*  We need to find the root DSE (or treename) if we haven't already. */
    if (!GetLDAPServerInfo(ldap)) {
        return (MDBHandle)0;
    }

    if (strchr(Object, '=')) {
        /*  Its an LDAP dn.  Don't try to convert it.  */
        ldapuname = strdup(Object);
    } else {
        MDBHandle handle = NULL;
        MDBValueStruct *V;

        if (MDBldap.mdbhandle) {
            V = MDBLDAPCreateValueStruct(MDBldap.mdbhandle, NULL);
        } else {
            /*  Use Public  */
            handle = MDBLDAPAuthenticate(NULL, NULL, "");
            V = MDBLDAPCreateValueStruct(handle, NULL);
        }
        ldapuname = MDBLDAPFindObject(Object, NULL, V);

        MDBLDAPDestroyValueStruct(V);
        if (handle) {
            MDBLDAPRelease(handle);
        }
    }

    /*  Now try to authenticate to the db with that uname and pwd.  */
    if (ldapuname) {
        if (ldapuname[0] != '\0') {
            err = ldap_simple_bind_s(ldap, ldapuname, Password);

            if (err==LDAP_SUCCESS) {
                context = MDBLDAPAddHandle(ldap);
                free(ldapuname);
                return (MDBHandle)(context);
            }
        }
    }
    free(ldapuname);
    return NULL;
}

BOOL
MDBLDAPRelease(MDBHandle Handle)
{
    if (MDBLDAPFindHandle(Handle)) {
        MDBLDAPRemoveHandle(Handle);
    }
    return TRUE;
}

MDBValueStruct*
MDBLDAPCreateValueStruct(MDBHandle Handle, const unsigned char *Context)
{
    MDBValueStruct *v = NULL;
    MDBContextStruct *context = NULL;

    if (Context) {
    if (strcmp(Context, "\\\\Internet Services\\Pats Messaging Server")==0) {
        int x=0;
        x++;
    }
    }
    if ((context = MDBLDAPFindHandle(Handle))) {
        v = malloc(sizeof(MDBValueStruct));
        if (v) {
            memset(v, 0, sizeof(MDBValueStruct));

            v->BaseDN = malloc(XPL_MAX_PATH + 1);
            v->context = context;

            if (!Context || (strcmp(Context, "\\")==0) || (strcmp(Context, "") == 0)) {
                if (MDBldap.dirType == MDBLDAP_TYPE_EDIR) {
                    strcpy(v->BaseDN, "");
                } else {
                    strcpy(v->BaseDN, MDBldap.rootDSE);
                }
                v->Flags = MDB_FLAGS_CONTEXT_VALID;
            } else {
                /*  here we go searching ( i have to set this so that i can search ) */
                v->Flags |= MDB_FLAGS_CONTEXT_VALID;
                v->BaseDN = MDBLDAPFindObject(Context, "", v);
                v->Flags = MDB_FLAGS_BASEDN_CHANGED;
                if (strlen(v->BaseDN) > 0) {
                    v->Flags |= MDB_FLAGS_CONTEXT_VALID;
                } else {
                    v->Flags ^= MDB_FLAGS_CONTEXT_VALID;
                }
            }
        }
    }

    return(v);
}

BOOL
MDBLDAPDestroyValueStruct(MDBValueStruct *V)
{
    if (V && MDB_CONTEXT_IS_VALID(V)) {
        if (!(V->Flags & MDB_FLAGS_CONTEXT_DUPLICATE)) {
            free(V->BaseDN);
            V->BaseDN = NULL;
        }

        if (V->Allocated) {
            MDBLDAPFreeValues(V);

            if (V->Value) {
                free(V->Value);
            }
        }

        free(V);
        return(TRUE);
    }

    return(FALSE);
}

MDBValueStruct*
MDBLDAPShareContext(MDBValueStruct *V)
{
    MDBValueStruct *n = NULL;

    if (V && MDB_CONTEXT_IS_VALID(V)) {
        n = malloc(sizeof(MDBValueStruct));
        if (n) {
            memset(n, 0, sizeof(MDBValueStruct));

            n->Flags = MDB_FLAGS_CONTEXT_DUPLICATE | V->Flags;
            n->BaseDN = V->BaseDN;
            n->context = V->context;
        }
    }

    return(n);
}

BOOL
MDBLDAPSetValueStructContext(const unsigned char *Context, MDBValueStruct *V)
{
    if (V && MDB_CONTEXT_IS_VALID(V)) {
        if (!Context || (strcmp(Context, "\\")==0) || (strcmp(Context, "") == 0)) {
            strcpy(V->BaseDN, "");
        } else {
            /*  here we go searching */
            V->BaseDN = MDBLDAPFindObject(Context, "", V);
            if (strlen(V->BaseDN) == 0) {
                V->Flags ^= MDB_FLAGS_CONTEXT_VALID;
            }
        }

        V->Flags |= MDB_FLAGS_BASEDN_CHANGED;
        return(TRUE);
    }

    return(FALSE);
}

MDBEnumStruct*
MDBLDAPCreateEnumStruct(MDBValueStruct *V)
{
    MDBEnumStruct *es = NULL;

    if (V && MDB_CONTEXT_IS_VALID(V)) {
        es = malloc(sizeof(MDBEnumStruct));
        if (es) {
            es->Initialized = FALSE;
            es->Entry = 0;
            es->V = NULL;
        }
    }

    return(es);
}

BOOL
MDBLDAPDestroyEnumStruct(MDBEnumStruct *E, MDBValueStruct *V)
{
    if (E) {
        if (E->Initialized) {
            if (E->V) {
                MDBLDAPDestroyValueStruct(E->V);
            }
        }

        free(E);
        return(TRUE);    
    }

    return(FALSE);
}

BOOL
MDBLDAPAddValue(const unsigned char *Value, MDBValueStruct *V)
{
    register unsigned char *ptr;

    if (Value && (*Value != '\0')) {
        if ((V->Used + 1) > V->Allocated) {
            ptr = (unsigned char *)realloc(V->Value, (V->Allocated + VALUE_ALLOC_SIZE) * sizeof(unsigned char *));
            if (ptr) {
                V->Value = (unsigned char **)ptr;
                V->Allocated += VALUE_ALLOC_SIZE;
            } else {
                if (V->Allocated) {
                    MDBLDAPFreeValues(V);

                    if (V->Value) {
                        free(V->Value);
                    }
                }

                V->Value = NULL;

                V->Used = 0;
                V->Allocated = 0;

                return(FALSE);
            }
        }

        ptr = strdup(Value);
        if (ptr) {
            V->Value[V->Used] = ptr;
            V->Used++;

            return(TRUE);
        }
    }

    return(FALSE);
}

BOOL
MDBLDAPFreeValue(unsigned long Index, MDBValueStruct *V)
{
    if (Index < V->Used) {
        free(V->Value[Index]);

        if (Index < (V->Used - 1)) {
            memmove(&V->Value[Index], &V->Value[Index + 1], ((V->Used - 1) - Index) * sizeof(unsigned char *));
        }

        V->Used--;
        return(TRUE);
    }

    return(FALSE);
}

BOOL
MDBLDAPFreeValues(MDBValueStruct *V)
{
    register unsigned long i;

    if (V->Allocated) {
        for (i = 0; i < V->Used; i++) {
            free(V->Value[i]);
        }

        if (V->Allocated > VALUE_ALLOC_SIZE) {
            if (V->Value) {
                free(V->Value);
            }

            V->Allocated = 0;
            V->Value = NULL;
        }

        V->Used = 0;
    }

    return(TRUE);
}

/*
    See the comments to MDBLDAP_Read()
*/
long
MDBLDAPRead(const unsigned char *Object, const unsigned char *Attribute, MDBValueStruct *V)
{
    return MDBLDAP_Read(Object, Attribute, V, FALSE);
}

/*
    Return one attribute value at a time for the particular Attribute of Object,
    or NULL when there are no more values.
    There is a tradeoff here:
    In order to improve speed (reduce wire traffic), all attribute values are
    retrieved at once, but returned one at a time to the user.  If there are a
    lot of values, this can be a major memory hog.  A possible improvement,
    if the DB supports it, might be to read some maximum-sized batch at a time.
    E is private, so we must take care to free, destroy or reset any of its
    variables.
    V should not be altered - just used for context (or authentication).

*/
const unsigned char*
MDBLDAPReadEx(const unsigned char *Object, const unsigned char *Attribute, MDBEnumStruct *E, MDBValueStruct *V)
{
    if (E) {
        if (!E->Initialized) {
            E->Entry = 0;

            if (Attribute && V && MDB_CONTEXT_IS_VALID(V)) {
                MDBValueStruct *tmpV = MDBLDAPShareContext(V);
                E->V = tmpV;
                MDBLDAPRead(Object, Attribute, tmpV);
                E->Initialized = TRUE;
            }
        }

        if ((E->V->Used > 0) && (E->Entry != (E->V->Used))) {
            return E->V->Value[E->Entry++];
        }

        E->Initialized = FALSE;
        if (E->V) {
            MDBLDAPDestroyValueStruct(E->V);
        }
    }

    return(NULL);
}

/*
    See the comments to MDBLDAP_Read()
*/
long
MDBLDAPReadDN(const unsigned char *Object, const unsigned char *Attribute, MDBValueStruct *V)
{
    return MDBLDAP_Read(Object, Attribute, V, TRUE);
}

/*
    This function is not documented in the mdb or ddb docs.  Its function is unknown.
*/
BOOL
MDBLDAPWriteTyped(const unsigned char *Object, const unsigned char *Attribute, const int AttrType, MDBValueStruct *V)
{
    return FALSE;
}

/*
    Write replaces current attributes on an object with those contained in V
*/
BOOL
MDBLDAPWrite(const unsigned char *Object, const unsigned char *Attribute, MDBValueStruct *V)
{
    return MDBLDAP_Write(Object, Attribute, V, LDAP_MOD_REPLACE, FALSE);
}

/*
    WriteDN replaces current attributes on an object with those contained in V
    saving DNs as absolute dns
*/
BOOL
MDBLDAPWriteDN(const unsigned char *Object, const unsigned char *Attribute, MDBValueStruct *V)
{
    return MDBLDAP_Write(Object, Attribute, V, LDAP_MOD_REPLACE, TRUE);
}

/*
    Clear delete all an attributes values
*/
BOOL
MDBLDAPClear(const unsigned char *Object, const unsigned char *Attribute, MDBValueStruct *V)
{
    return MDBLDAP_Write(Object, Attribute, V, LDAP_MOD_DELETE, FALSE);
}

/*
    Add adds a single value to an attribute
*/
BOOL
MDBLDAPAdd(const unsigned char *Object, const unsigned char *Attribute, const unsigned char *Value, MDBValueStruct *V)
{
    MDBValueStruct *V2;
    BOOL result;
    V2 = MDBLDAPShareContext(V);
    MDBLDAPAddValue(Value, V2);
    result = MDBLDAP_Write(Object, Attribute, V2, LDAP_MOD_ADD, FALSE);
    MDBLDAPDestroyValueStruct(V2);
    return result;
}

/*
    Adds a single value to an attribute
    saving DNs as absolute DNs
*/
BOOL
MDBLDAPAddDN(const unsigned char *Object, const unsigned char *Attribute, const unsigned char *Value, MDBValueStruct *V)
{
    MDBValueStruct *V2;
    BOOL result;
    V2 = MDBLDAPShareContext(V);
    MDBLDAPAddValue(Value, V2);
    result = MDBLDAP_Write(Object, Attribute, V2, LDAP_MOD_ADD, TRUE);
    MDBLDAPDestroyValueStruct(V2);
    return result;
}

/*
    Removes a single value from an attribute
*/
BOOL
MDBLDAPRemove(const unsigned char *Object, const unsigned char *Attribute, const unsigned char *Value, MDBValueStruct *V)
{
    MDBValueStruct *V2;
    BOOL result;
    V2 = MDBLDAPShareContext(V);
    MDBLDAPAddValue(Value, V2);
    result = MDBLDAP_Write(Object, Attribute, V2, LDAP_MOD_DELETE, FALSE);
    MDBLDAPDestroyValueStruct(V2);
    return result;
}

/*
    Removes a single DN value from an attribute
*/
BOOL
MDBLDAPRemoveDN(const unsigned char *Object, const unsigned char *Attribute, const unsigned char *Value, MDBValueStruct *V)
{
    MDBValueStruct *V2;
    BOOL result;
    V2 = MDBLDAPShareContext(V);
    MDBLDAPAddValue(Value, V2);
    result = MDBLDAP_Write(Object, Attribute, V2, LDAP_MOD_DELETE, TRUE);
    MDBLDAPDestroyValueStruct(V2);
    return result;
}

BOOL
MDBLDAPIsObject(const unsigned char *Object, MDBValueStruct *V)
{
    char *dn;
    BOOL result = FALSE;

    if (V && MDB_CONTEXT_IS_VALID(V)) {
        dn = MDBLDAPFindObject(Object, V->BaseDN, V);
        if (strlen(dn) > 0) {
            result = TRUE;
        }
        free(dn);
    }

    return result;
}

BOOL
MDBLDAPGetObjectDetails(const unsigned char *Object, unsigned char *Type, unsigned char *RDN, unsigned char *DN, MDBValueStruct *V)
{
    int err;
    LDAPMessage *pres = NULL;
    LDAPMessage *pres2 = NULL;
    char *dn;
    char *obj;
    char *ptr;
    BOOL result = FALSE;

    if (!V || !MDB_CONTEXT_IS_VALID(V) || !translationtable) {
        return result;
    }

    /* zip past leading backslashes */
    ptr = (char *) Object;
    while (*ptr == '\\') {
        ptr++;
    }

    if (MDBldap.dirType == MDBLDAP_TYPE_EDIR &&
          (*ptr == '\0' || strcmp(ptr, MDBldap.treeName) == 0)) {
        if (Type) {
            strcpy(Type, "Top");
        }

        if (DN) {
            sprintf(DN, "\\%s", MDBldap.treeName);
        }

        if (RDN) {
            sprintf(RDN, "\\%s", MDBldap.treeName);
        }

        return TRUE;
    }

    obj = MDBLDAPFindObject(Object, V->BaseDN, V);
    if (strlen(obj) > 0) {
        XplWaitOnLocalSemaphore(V->context->sem);
        err = ldap_search_s(V->context->bindhandle, obj, LDAP_SCOPE_BASE, "(objectClass=*)", ObjectClassAttributes, 0, &pres);
        if (err==LDAP_SUCCESS) {
            pres2 = ldap_first_entry(V->context->bindhandle, pres);
            if (pres2) {
                dn = ldap_get_dn(V->context->bindhandle, pres2);
                if (DN) {
                    /*  This needs to change for non-eDir...                 */
                    DN += sprintf(DN, "\\%s", MDBldap.treeName);
                    MDBLDAPldap2mdb(dn, DN, NULL);
                }

                if (RDN) {
                    MDBLDAPldap2mdb(dn, RDN, V->BaseDN);
                }
                ldap_memfree(dn);

                if (Type) {
                    MDBValueStruct *objclass;
                    unsigned long class;
                    char *type;
                
                    XplSignalLocalSemaphore(V->context->sem);
                    objclass = MDBLDAPShareContext(V);
                
                    MDBLDAP_Read(Object, "objectClass", objclass, FALSE);

                    /* now find the mdb class and set the type */
                    for(class=0;class<objclass->Used;class++) {
                        if (class == 0) {
                            /* In case we don't find a match */
                            strcpy(Type, objclass->Value[class]);
                        }

                        if ((type = FindMDBClass(objclass->Value[class]))) {
                            strcpy(Type, type);
                            break;
                        }
                    }

                    MDBLDAPDestroyValueStruct(objclass);
                    XplWaitOnLocalSemaphore(V->context->sem);
                }
                result=TRUE;
            }
        }
        if (pres) {
            ldap_msgfree(pres);
        }
        XplSignalLocalSemaphore(V->context->sem);
    }

    free(obj);
    return result;
}

BOOL
MDBLDAPVerifyPassword(const unsigned char *Object, const unsigned char *Password, MDBValueStruct *V)
{
    LDAP *ldap;
    char *dn;
    int err;
    BOOL result = FALSE;
    int version = LDAP_VERSION3;

    if (V && MDB_CONTEXT_IS_VALID(V)) {
        dn = MDBLDAPFindObject(Object, V->BaseDN, V);
        if (strlen(dn) > 0) {
            /*  Need to set this here? It was done in Authenticate(), but   */
            /*  if this can be called first, we must do this here also.     */
            ldap_set_option(NULL, LDAP_OPT_PROTOCOL_VERSION, &version);

            if ((ldap = ldap_init(ldapServer, ldapPort))) {
                /*  Now try to authenticate to the db with that uname & pwd */
                err = ldap_simple_bind_s(ldap, dn, Password);
                if (err==LDAP_SUCCESS) {
                    result=TRUE;
                }
                ldap_unbind_s(ldap);
            }
        }
        free(dn);
    }
    return result;
}

BOOL
MDBLDAPVerifyPasswordEx(const unsigned char *Object, const unsigned char *Password, MDBValueStruct *V)
{
    return MDBLDAPVerifyPassword(Object, Password, V);
}

BOOL
MDBLDAPChangePassword(const unsigned char *Object, const unsigned char *OldPassword, const unsigned char *NewPassword, MDBValueStruct *V)
{
    MDBValueStruct *V2;
    BOOL result;

    V2 = MDBLDAPShareContext(V);
    MDBLDAPAddValue(NewPassword, V2);
    result = MDBLDAP_Write(Object, "userpassword", V2, LDAP_MOD_REPLACE, FALSE);
    MDBLDAPDestroyValueStruct(V2);
    return result;
}

BOOL
MDBLDAPChangePasswordEx(const unsigned char *Object, const unsigned char *OldPassword, const unsigned char *NewPassword, MDBValueStruct *V)
{
    return MDBLDAPChangePassword(Object, OldPassword, NewPassword, V);
}

long
MDBLDAPEnumerateObjects(const unsigned char *Container, const unsigned char *Type, const unsigned char *Pattern, MDBValueStruct *V)
{
    char *container = NULL, *type=NULL, *objectdn;
    char *ldapclass=NULL;
    int err, num=0;
    LDAPMessage *pres = NULL;
    LDAPMessage *pres2 = NULL;
    char tmpValue[MDB_MAX_OBJECT_CHARS + 1];

    if (!Container || !translationtable)
        return -1;

    if (strcmp(Container, "\\")==0) {
        container = strdup("");
    } else {
        container = MDBLDAPFindObject(Container, V->BaseDN, V);
    }

    if (Type && (ldapclass = FindLDAPClass(Type))) {
        type = malloc((strlen(ldapclass)+15)*sizeof(unsigned char)); /* 14 for (objectClass=) and one for null */
        sprintf(type, "(objectClass=%s)", ldapclass);
    } else {
        type=strdup("(objectClass=*)");
    }

    /* do an ldap search in Container for all objects */
    XplWaitOnLocalSemaphore(V->context->sem);
    err = ldap_search_s(V->context->bindhandle, container, LDAP_SCOPE_ONELEVEL, type, DNAttributes, 0, &pres);
    if (err==LDAP_SUCCESS) {
        pres2 = ldap_first_entry(V->context->bindhandle, pres);
        while (pres2) {
            objectdn = ldap_get_dn(V->context->bindhandle, pres2);
            MDBLDAPldap2mdb(objectdn, tmpValue, NULL);
            MDBLDAPAddValue(tmpValue, V);
            num++;
            ldap_memfree(objectdn);
            pres2 = ldap_next_entry(V->context->bindhandle, pres2);
        }
    }

    if (pres) { 
        ldap_msgfree(pres);
    }

    XplSignalLocalSemaphore(V->context->sem);
    free(type);
    free(container);
    return num;
}

const unsigned char*
MDBLDAPEnumerateObjectsEx(const unsigned char *Container, const unsigned char *Type, const unsigned char *Pattern, unsigned long Flags, MDBEnumStruct *E, MDBValueStruct *V)
{

    if (E) {
        if (!E->Initialized) {
            E->Entry = 0;

            if (Container && V && MDB_CONTEXT_IS_VALID(V)) {
                E->V = V;
                MDBLDAPFreeValues(E->V);
                MDBLDAPEnumerateObjects(Container, Type, Pattern, V);
                E->Initialized = TRUE;
            }
        }

        if (E->Entry < E->V->Used) {
            return E->V->Value[E->Entry++];
        }

        E->Initialized = FALSE;
    }

    return(NULL);
}

BOOL
MDBLDAPAddAddressRestriction(const unsigned char *Object, const unsigned char *Server, MDBValueStruct *V)
{
    return TRUE;
}

/* this code is commented out pending review of its necessity*/
const unsigned char*
MDBLDAPEnumerateAttributesEx(const unsigned char *Object, MDBEnumStruct *E, MDBValueStruct *V)
{
/*
    int err;
    LDAPMessage *pres = NULL;
    LDAPMessage *pres2 = NULL;
    char *attribute=NULL;
    BerElement *berptr=NULL;
    char *obj;
    unsigned char *result = NULL;

    if (!V || !MDB_CONTEXT_IS_VALID(V)) {
        return result;
    }

    obj = MDBLDAPFindObject(Object, V->BaseDN, V);
    if (strlen(obj) > 0) {
        XplWaitOnLocalSemaphore(V->context->sem);
        err = ldap_search_s(V->context->bindhandle, obj, LDAP_SCOPE_BASE, "(objectClass=*)", NULL, 1, &pres);
        if (err==LDAP_SUCCESS) {
            pres2 = ldap_first_entry(V->context->bindhandle, pres);
            if (pres2) {
                attribute = ldap_first_attribute(V->context->bindhandle, pres2, &berptr);
                while (attribute) {
                    MDBLDAPAddValue(attribute, V);
                    ldap_memfree(attribute);
                    attribute = ldap_next_attribute(V->context->bindhandle, pres2, berptr);
                }
                if (berptr)
                    ber_free(berptr, 0);
            }
        }
        if (pres) {
            ldap_msgfree(pres);
        }
        XplSignalLocalSemaphore(V->context->sem);
    }

    free(obj);
    return result;
*/
    if (V) {
        V->ErrNo = ERR_NOT_IMPLEMENTED;
    }

    return NULL;
}

BOOL
MDBLDAPCreateObject(const unsigned char *Object, const unsigned char *Class, MDBValueStruct *Attribute, MDBValueStruct *Data, MDBValueStruct *V)
{       
    BOOL multi;
    unsigned long end;
    char *tmp;
#ifndef WIN32
    fprintf(outfile, "CreateObject(%s, %s)\n", Object, Class); /* weird Windows compile error */
#endif

    char objname[MDB_MAX_OBJECT_CHARS + 1];
    unsigned char *object;
    unsigned char *o;
    unsigned long x, y, i;
    int err;
    LDAPMod **attrs;
    char *objclass;
    char *tmpclass;

    if (!Object || !Class || !V || !translationtable) {
        return FALSE;
    }

    /* Create a local copy to work with */
    object = strdup(Object);

    /* convert Object string into ldapcode */
    objclass = MDBLDAPFindNaming(Class);

    o = strrchr(object, '\\');
    if (o) {
        unsigned char *a;

        *o = '\0';
        o++;

        a = MDBLDAPFindObject(object, V->BaseDN, V);
        sprintf(objname, "%s=%s,%s", objclass, o, a);

        free(a);
    } else if (V->BaseDN && V->BaseDN[0]) {
        sprintf(objname, "%s=%s,%s", objclass, object, V->BaseDN);
    } else {
        sprintf(objname, "%s=%s", objclass, object);
    }

    free(object);

    /*  Set up the array of attributes.  If there are duplicates,            */
    /*  (which is rare) we won't use all of them, but that's OK.             */
    attrs = malloc((Attribute->Used+3) * sizeof(LDAPMod *));

    /*  Loop thru the Attributes, adding each to the LDAPMod list.           */
    /*  eDirectory allows for listing the same attribute multiple times      */
    /*  in the list, but OpenLDAP does not.  So, we need to determine if     */
    /*  there are multiples, and combine them into one attribute entry,      */
    /*  having multiple values.                                              */
    for (x = 0, i=0 ; x < Attribute->Used ; x++) {
        multi = FALSE;
        end   = 0;

        for (y = 0 ; y < x ; y++) {
            if (XplStrCaseCmp(Attribute->Value[x], Attribute->Value[y]) == 0) {
                multi = TRUE;
                for (end = 0 ; attrs[y]->mod_bvalues[end] != NULL ; end++) {
                    ;    /*  just find the ending NULL                       */
                }
                break;               /*  found a duplicate                   */
            }
        }

        if (!multi) {
            attrs[i] = malloc(sizeof(LDAPMod));
            /*  mod_op should be set based on the syntax type.               */
            attrs[i]->mod_op = LDAP_MOD_BVALUES | LDAP_MOD_ADD;
            attrs[i]->mod_type = strdup(&Attribute->Value[x][1]);
            RemoveSpaces(attrs[i]->mod_type);
            attrs[i]->mod_bvalues = malloc(2*sizeof(struct berval *));
            attrs[i]->mod_bvalues[end] = malloc(sizeof(struct berval));

            if (Attribute->Value[x][0] == MDB_ATTR_SYN_DIST_NAME) {
                tmp = MDBLDAPFindObject(Data->Value[x], V->BaseDN, V);
                attrs[i]->mod_bvalues[end]->bv_len = strlen(tmp);
                attrs[i]->mod_bvalues[end]->bv_val = tmp;
            } else {
                attrs[i]->mod_bvalues[end]->bv_len = strlen(Data->Value[x]);
                attrs[i]->mod_bvalues[end]->bv_val = strdup(Data->Value[x]);
            }
            attrs[i]->mod_bvalues[end+1] = NULL;
            i++;
        } else {
            if (end >= 0) {
                struct berval *new = realloc(attrs[y]->mod_bvalues,
                                            (end+1+1)*sizeof(struct berval *));
                if (new) {
                    attrs[y]->mod_bvalues = (struct berval **)new;
                } else {
                    /*  out of memory - set ErrNo and return FALSE           */
                    ldap_mods_free(attrs, 1);
                    return FALSE;
                }
                attrs[y]->mod_bvalues[end] = malloc(sizeof(struct berval));

                if (Attribute->Value[x][0] == MDB_ATTR_SYN_DIST_NAME) {
                    tmp = MDBLDAPFindObject(Data->Value[x], V->BaseDN, V);
                    attrs[y]->mod_bvalues[end]->bv_len = strlen(tmp);
                    attrs[y]->mod_bvalues[end]->bv_val = tmp;
                } else {
                    attrs[y]->mod_bvalues[end]->bv_len = strlen(Data->Value[x]);
                    attrs[y]->mod_bvalues[end]->bv_val = strdup(Data->Value[x]);
                }
                attrs[y]->mod_bvalues[end+1] = NULL;
            } else {
                /*  something is really wrong                             */
            }
        }
    }

    /* now add the objectclass */
    attrs[i] = malloc(sizeof(LDAPMod));
    attrs[i]->mod_op = LDAP_MOD_BVALUES | LDAP_MOD_ADD;
    attrs[i]->mod_type = strdup("objectClass");
    attrs[i]->mod_bvalues = malloc(2*sizeof(struct berval *));
    attrs[i]->mod_bvalues[0] = malloc(sizeof(struct berval));
    tmpclass = strdup(Class);
    RemoveSpaces(tmpclass);
    attrs[i]->mod_bvalues[0]->bv_val = tmpclass;
    attrs[i]->mod_bvalues[0]->bv_len = strlen(tmpclass);
    attrs[i]->mod_bvalues[1] = NULL;

    /*  and finally add the object class type as an attribute  */
    i++;
    attrs[i] = malloc(sizeof(LDAPMod));
    attrs[i]->mod_op = LDAP_MOD_BVALUES | LDAP_MOD_ADD;
    attrs[i]->mod_type = strdup(objclass);
    attrs[i]->mod_bvalues = malloc(2*sizeof(struct berval *));
    attrs[i]->mod_bvalues[0] = malloc(sizeof(struct berval));
    tmpclass = strdup(Object);
    /* RemoveSpaces(tmpclass);     object names can have spaces */
    attrs[i]->mod_bvalues[0]->bv_val = tmpclass;
    attrs[i]->mod_bvalues[0]->bv_len = strlen(tmpclass);
    attrs[i]->mod_bvalues[1] = NULL;

    /* null terminate the array */
    attrs[i+1] = NULL;

    XplWaitOnLocalSemaphore(V->context->sem);
    err = ldap_add_s(V->context->bindhandle, objname, attrs);
    XplSignalLocalSemaphore(V->context->sem);
    ldap_mods_free(attrs, 1);
    if ((err != LDAP_SUCCESS) && (err != LDAP_ALREADY_EXISTS)) {
        printf("ldap_add_s() failed with %s\n", ldap_err2string(err));  
        return FALSE;
    }

    return TRUE;
}

BOOL
MDBLDAPDeleteObject(const unsigned char *Object, BOOL Recursive, MDBValueStruct *V)
{
    char *dn;
    int err;
    BOOL result=FALSE;

    if (V && MDB_CONTEXT_IS_VALID(V)) {
        dn=MDBLDAPFindObject(Object, V->BaseDN, V);
        if (strlen(dn) > 0) {
            XplWaitOnLocalSemaphore(V->context->sem);
            err = ldap_delete_s(V->context->bindhandle, dn);
            XplSignalLocalSemaphore(V->context->sem);
            if (err == LDAP_SUCCESS) {
                result = TRUE;
            }
        }
        free(dn);
    }
    return result;
}

BOOL
MDBLDAPRenameObject(const unsigned char *ObjectOld, const unsigned char *ObjectNew, MDBValueStruct *V)
{
    if (V) {
        V->ErrNo = ERR_NOT_IMPLEMENTED;
    }
    return FALSE;
}

BOOL
MDBLDAPDefineAttribute(const unsigned char *Attribute, const unsigned char *ASN1, unsigned long Type, BOOL SingleValue, BOOL ImmediateSync, BOOL Public, MDBValueStruct *V)
{
    /* since this isn't possible in ldap, return true */
    fprintf(outfile, "DefineAttribute\tAttribute=%s\tASN1=%s\tType=%ld\tSingleValue=%d\tImmediateSync=%d\tPublic=%d\n", Attribute, ASN1, Type, SingleValue, ImmediateSync, Public);
    return TRUE;
}

BOOL
MDBLDAPUndefineAttribute(const unsigned char *Attribute, MDBValueStruct *V)
{
    /* since this isn't possible in ldap, return true */
    fprintf(outfile, "UnDefineAttribute\tAttribute=%s\n", Attribute);
    return TRUE;
}

BOOL
MDBLDAPDefineClass(const unsigned char *Class, const unsigned char *ASN1, BOOL Container, MDBValueStruct *Superclass, MDBValueStruct *Containment, MDBValueStruct *Naming, MDBValueStruct *Mandatory, MDBValueStruct *Optional, MDBValueStruct *V)
{
    /* since this isn't possible in ldap, return true */
    unsigned long x, len=0;
    char super[5000];
    char containment[5000];
    char naming[5000];
    char mandatory[5000];
    char optional[5000];

    /* empty out all the vars */
    memset(super, '\0', 5000);
    memset(containment, '\0', 5000);
    memset(naming, '\0', 5000);
    memset(mandatory, '\0', 5000);
    memset(optional, '\0', 5000);

    for(x=0;x<Superclass->Used;x++) {
        len += sprintf(super+len, "%s-", Superclass->Value[x]);
    }

    for(x=0,len=0;x<Containment->Used;x++) {
        len += sprintf(containment+len, "%s-", Containment->Value[x]);
    }

    for(x=0,len=0;x<Naming->Used;x++) {
        len += sprintf(naming+len, "%s-", Naming->Value[x]);
    }

    for(x=0,len=0;x<Mandatory->Used;x++) {
        len += sprintf(mandatory+len, "%s-", Mandatory->Value[x]);
    }

    for(x=0,len=0;x<Optional->Used;x++) {
        len += sprintf(optional+len, "%s-", Optional->Value[x]);
    }

    fprintf(outfile, "DefineClass\tClass=%s\tASN1=%s\tContainer=%d\tsuper=%s\tcontainment=%s\tnaming=%s\tmandatory=%s\toptional=%s\n", Class, ASN1, Container, super, containment, naming, mandatory, optional);
    return TRUE;
}

BOOL
MDBLDAPAddAttribute(const unsigned char *Attribute, const unsigned char *Class, MDBValueStruct *V)
{
    /* since this isn't possible in ldap, return true */
    fprintf(outfile, "AddAttribute\tAttribute=%s\tClass=%s\n", Attribute, Class);
    return TRUE;
}

BOOL
MDBLDAPUndefineClass(const unsigned char *Class, MDBValueStruct *V)
{
    /* since this isn't possible in ldap, return true */
    fprintf(outfile, "UndefineClass\tClass=%s\n", Class);
    return TRUE;
}

static __inline void
addSubClasses(MDBldapClassStruct *class, MDBValueStruct *V)
{
    if (class->subClasses) {
        unsigned long i = 0;

        while (class->subClasses[i]) {
            unsigned long n = 0;

            /* Containment is inherited, unless overwritten */
            if (class->subClasses[i]->inheritContainment) {
                while (class->subClasses[i]->names[n]) {
                    MDBLDAPAddValue(class->subClasses[i]->names[n], V);
                    n++;
                }

                if (class->subClasses[i]->names[0]) {
                    addSubClasses(class->subClasses[0], V);
                }
            }

            i++;
        }
    }
}

static __inline BOOL
MDBLDAPListContainableClassesByClass(const unsigned char *type, MDBValueStruct *V)
{
    unsigned long i = 0;
    MDBldapClassStruct *class = NULL;

    /* Find this type in the cached schema */
    while (class == NULL && MDBldap.schema[i].names) {
        unsigned long n = 0;

        while (class == NULL && MDBldap.schema[i].names[n]) {
            if (strcmp(MDBldap.schema[i].names[n], type) == 0) {
                class = &MDBldap.schema[i];
            }
            n++;
        }

        i++;
    }

    if (class) {
        if (!class->leafOnly) {
            if (class->containableClasses) {
                i = 0;
                while (class->containableClasses[i]) {
                    unsigned long n = 0;

                    while (class->containableClasses[i]->names[n]) {
                        MDBLDAPAddValue(class->containableClasses[i]->names[n], V);

                        addSubClasses(class->containableClasses[i], V);
                        n++;
                    }

                    i++;
                }
            } else {
                /*
                    It isn't a leaf only, and there are no known containable classes, so
                    assume that the LDAP server does not report containment details, and we
                    have to list all known classes
                */
                i = 0;
                while (MDBldap.schema[i].names) {
                    unsigned long n = 0;

                    while (MDBldap.schema[i].names[n]) {
                        MDBLDAPAddValue(MDBldap.schema[i].names[n], V);
                        n++;
                    }
                    i++;
                }
            }
        }
        return(TRUE);
    }

    return FALSE;
}

BOOL
MDBLDAPListContainableClasses(const unsigned char *Container, MDBValueStruct *V)
{
    unsigned char type[MDB_MAX_OBJECT_CHARS + 1];

    if (MDBLDAPGetObjectDetails(Container, type, NULL, NULL, V)) {
        return(MDBLDAPListContainableClassesByClass(type, V));
    } else {
        return(FALSE);
    }
}

const unsigned char*
MDBLDAPListContainableClassesEx(const unsigned char *Container, MDBEnumStruct *E, MDBValueStruct *V)
{
    fprintf(outfile, "ListContainableClassesEx\tContainer=%s\n", Container);
    if (V) {
        V->ErrNo = ERR_NOT_IMPLEMENTED;
    }

    return NULL;
}

BOOL
MDBLDAPGrantObjectRights(const unsigned char *Object, const unsigned char *TrusteeDN, BOOL Read, BOOL Write, BOOL Delete, BOOL Rename, BOOL Admin, MDBValueStruct *V)
{
    int     err, privileges = 0;
    BOOL    result = FALSE;
    char    *truster, *trustee;
    LDAPMod **attrs;
    char    aclValue[MDB_MAX_OBJECT_CHARS
                           + 7    /*  strlen("#entry#")                      */
                           + 15   /*  strlen("#[Entry Rights]")              */
                           + 1];  /*  '\0'                                   */

    /*  If we're not using eDirectory, then get out of here -                */
    /*  until we have this implemented for OpenLDAP and/or others...         */
    if (MDBldap.dirType != MDBLDAP_TYPE_EDIR) {
        fprintf(outfile, "GrantObjectRights(%s, %s, %d, %d, %d, %d, %d)\n",
                Object, TrusteeDN, Read, Write, Delete, Rename, Admin);
        return TRUE;
    }

    /*  The following works on eDirectory only.                              */
    if (V && (V->Flags & MDB_CONTEXT_IS_VALID(V))
                       && Object && (Object[0] != '\0')
                       && TrusteeDN && (TrusteeDN[0] != '\0')) {
        truster = MDBLDAPFindObject(Object, V->BaseDN, V);
        trustee = MDBLDAPFindObject(TrusteeDN, V->BaseDN, V);

        /*  If the truster is the tree, or rootDSE, then it's OK             */
        /*  for FindObject to have returned "" (at least with eDir).         */
        /*  LDAP_DS_ENTRY_XXX constants were pulled from ldap_ds_constants.h */
        /*  which is from the Novell cldap library.                          */
        if (truster != NULL && trustee != NULL && trustee[0] != '\0') {
            /*  Bit values for [Entry Rights] of ACL.                        */
            privileges |= (Read)?   0x00000001L : 0; /* LDAP_DS_ENTRY_BROWSE */
            privileges |= (Write)?  0x00000002L : 0; /* LDAP_DS_ENTRY_ADD    */
            privileges |= (Delete)? 0x00000004L : 0; /* LDAP_DS_ENTRY_DELETE */
            privileges |= (Rename)? 0x00000008L : 0; /* LDAP_DS_ENTRY_RENAME */
            privileges |= (Admin)?  0x00000010L : 0; /*  "  "  " _SUPERVISOR */
            /* other possible       0x00000040L : 0;     "  "  "_INHERIT_CTL */
            sprintf(aclValue, "%d", privileges);
            strcat(aclValue, "#entry#");        /*  other choice = #subtree# */
            strcat(aclValue, trustee);
            /*  This is one of the 2 special attribute names in ACLs in eDir */
            /*  The other is "[All Attributes Rights]" (see dconst.h in      */
            /*  dclient).  We use that in MDBLDAPGrantAttributeRights below. */
            strcat(aclValue, "#[Entry Rights]");/* other = [All Attributes]  */

            /*  Set up the LDAPMod.  */
            attrs = malloc(2*sizeof(LDAPMod *));   /*  ACL plus ending null  */
            if (attrs) {
                attrs[0] = malloc(sizeof(LDAPMod));
                if (attrs[0]) {
                    attrs[0]->mod_op = LDAP_MOD_BVALUES | LDAP_MOD_ADD;
                    attrs[0]->mod_type = strdup(A_ACL); /*  for eDir,  "ACL" */
                    if (attrs[0]->mod_type) {
                        RemoveSpaces(attrs[0]->mod_type);
                        attrs[0]->mod_bvalues = malloc(2*sizeof(struct berval *));
                        if (attrs[0]->mod_bvalues) {
                            attrs[0]->mod_bvalues[0] = malloc(sizeof(struct berval));
                            if (attrs[0]->mod_bvalues[0]) {
                                attrs[0]->mod_bvalues[0]->bv_len = strlen(aclValue);
                                attrs[0]->mod_bvalues[0]->bv_val = aclValue;
                                attrs[0]->mod_bvalues[1] = NULL;
                                attrs[1] = NULL;

                                XplWaitOnLocalSemaphore(V->context->sem);
                                err = ldap_modify_s(V->context->bindhandle, truster, attrs);
                                XplSignalLocalSemaphore(V->context->sem);
                                if (err == LDAP_SUCCESS) {
                                    result = TRUE;
                                } else {
                                    printf("MDBLDAPGrantObjectRights() failed:\n");  
                                    printf("ldap_modify_s(): %s\n", ldap_err2string(err));  
                                    V->ErrNo = err;
                                }
                                free(attrs[0]->mod_bvalues[0]);
                            } else {
                                V->ErrNo = ERR_INSUFFICIENT_MEMORY;
                            }
                            free(attrs[0]->mod_bvalues);
                        } else {
                            V->ErrNo = ERR_INSUFFICIENT_MEMORY;
                        }
                        free(attrs[0]->mod_type);
                    } else {
                        V->ErrNo = ERR_INSUFFICIENT_MEMORY;
                    }
                    free(attrs[0]);
                } else {
                    V->ErrNo = ERR_INSUFFICIENT_MEMORY;
                }
                free(attrs);
            } else {
                V->ErrNo = ERR_INSUFFICIENT_MEMORY;
            }
        } else {
            V->ErrNo = ERR_NO_SUCH_ENTRY;
        }
        if (truster) {
            free(truster);
        }
        if (trustee) {
            free(trustee);
        }
    } else {
        if (V) {
            V->ErrNo = ERR_INVALID_REQUEST;
        }
    }

    return result;

}   /*  End  MDBLDAPGrantObjectRights()                                      */


BOOL
MDBLDAPGrantAttributeRights(const unsigned char *Object, const unsigned char *Attribute, const unsigned char *TrusteeDN, BOOL Read, BOOL Write, BOOL Admin, MDBValueStruct *V)
{
    int     err, privileges = 0;
    BOOL    result = FALSE;
    LDAPMod **attrs;
    char    *object, *trustee;
    char    aclValue[MDB_MAX_OBJECT_CHARS
                           + 7    /*  strlen("#entry#")                      */
                           + 24   /*  strlen("#[All Attributes Rights]")     */
                           + 1];  /*  '\0'                                   */

    /*  If we're not using eDirectory, then get out of here -                */
    /*  until we have this implemented for OpenLDAP and/or others...         */
    if (MDBldap.dirType != MDBLDAP_TYPE_EDIR) {
        fprintf(outfile, "GrantAttributeRights(%s, %s, %s, %d, %d, %d)\n",
                Object, Attribute, TrusteeDN, Read, Write, Admin);
        return TRUE;
    }

    /*  The following works on eDirectory only.                              */
    /*  Attribute might be NULL, in which case it means to                   */
    /*  grant [All Attributes Rights].                                       */
    if (V && (V->Flags & MDB_CONTEXT_IS_VALID(V))
                       && Object && (Object[0] != '\0')
                       && TrusteeDN && (TrusteeDN[0] != '\0')) {
        object = MDBLDAPFindObject(Object, V->BaseDN, V);
        trustee = MDBLDAPFindObject(TrusteeDN, V->BaseDN, V);

        /*  If the object is the tree, or rootDSE, then it's OK              */
        /*  for FindObject to have returned "" (at least with eDir).         */
        /*  LDAP_DS_ATTR_XXX constants were pulled from ldap_ds_constants.h  */
        /*  which is from the Novell cldap library.                          */
        if (object != NULL && trustee != NULL && trustee[0] != '\0') {
            /*  Bit values for [All Attributes Rights] of ACL and            */
            /*  individual attribute rights of ACL.                          */
            privileges |= (Read)?   0x00000001L |   /* LDAP_DS_ATTR_COMPARE  */
                                    0x00000002L : 0;/* LDAP_DS_ATTR_READ     */
            privileges |= (Write)?  0x00000004L : 0;/* LDAP_DS_ATTR_WRITE    */
            privileges |= (Admin)?  0x00000008L |   /* LDAP_DS_ATTR_SELF     */
                                    0x00000020L : 0;/*LDAP_DS_ATTR_SUPERVISOR*/
            /* other possible       0x00000040L : 0;    "   "  "_INHERIT_CTL */
            /* other possible       0x20000000L : 0;  LDAP_DS_DYNAMIC_ACL    */
            sprintf(aclValue, "%d", privileges);
            strcat(aclValue, "#entry#");        /*  other choice = #subtree# */
            strcat(aclValue, trustee);
            /*  This is one of the 2 special attribute names in ACLs in eDir */
            /*  The other is "[Entry Rights]" (see dconst.h in dclient).     */
            /*  We use that in MDBLDAPGrantObjectRights above.               */
            if (Attribute == NULL || Attribute[0] == '\0') {
                strcat(aclValue, "#[All Attributes Rights]");
            } else {
                strcat(aclValue, Attribute);
            }

            /*  Set up the LDAPMod.  */
            attrs = malloc(2*sizeof(LDAPMod *));   /*  ACL plus ending null  */
            if (attrs) {
                attrs[0] = malloc(sizeof(LDAPMod));
                if (attrs[0]) {
                    attrs[0]->mod_op = LDAP_MOD_BVALUES | LDAP_MOD_ADD;
                    attrs[0]->mod_type = strdup(A_ACL);/* for eDir, it's "ACL"*/
                    if (attrs[0]->mod_type) {
                        RemoveSpaces(attrs[0]->mod_type);
                        attrs[0]->mod_bvalues = malloc(2 * sizeof(struct berval *));
                        if (attrs[0]->mod_bvalues) {
                            attrs[0]->mod_bvalues[0] = malloc(sizeof(struct berval));
                            if (attrs[0]->mod_bvalues[0]) {
                                attrs[0]->mod_bvalues[0]->bv_len = strlen(aclValue);
                                attrs[0]->mod_bvalues[0]->bv_val = aclValue;
                                attrs[0]->mod_bvalues[1] = NULL;
                                attrs[1] = NULL;

                                XplWaitOnLocalSemaphore(V->context->sem);
                                err = ldap_modify_s(V->context->bindhandle, object, attrs);
                                XplSignalLocalSemaphore(V->context->sem);
                                if (err == LDAP_SUCCESS) {
                                    result = TRUE;
                                } else {
                                    printf("MDBLDAPGrantAttributeRights() failed:\n");  
                                    printf("ldap_modify_s(): %s\n", ldap_err2string(err));  
                                    V->ErrNo = err;
                                }
                                free(attrs[0]->mod_bvalues[0]);
                            } else {
                                V->ErrNo = ERR_INSUFFICIENT_MEMORY;
                            }
                            free(attrs[0]->mod_bvalues);
                        } else {
                            V->ErrNo = ERR_INSUFFICIENT_MEMORY;
                        }
                        free(attrs[0]->mod_type);
                    } else {
                        V->ErrNo = ERR_INSUFFICIENT_MEMORY;
                    }
                    free(attrs[0]);
                } else {
                    V->ErrNo = ERR_INSUFFICIENT_MEMORY;
                }
                free(attrs);
            } else {
                V->ErrNo = ERR_INSUFFICIENT_MEMORY;
            }
        } else {
            V->ErrNo = ERR_NO_SUCH_ENTRY;
        }
        if (object) {
            free(object);
        }
        if (trustee) {
            free(trustee);
        }
    } else {
        if (V) {
            V->ErrNo = ERR_INVALID_REQUEST;
        }
    }

    return result;

}   /*  End  MDBLDAPGrantAttributeRights()                                   */


static __inline MDBldapClassStruct*
findClassDefinition(unsigned char *name)
{
    unsigned long i = 0;

    while (MDBldap.schema[i].names != NULL) {
        unsigned long n = 0;

        while (MDBldap.schema[i].names[n]) {
            if (strcmp(MDBldap.schema[i].names[n], name) == 0) {
                return(&(MDBldap.schema[i]));
            }

            n++;
        }

        i++;
    }

    return(NULL);
}

#define addSchemaValue(result, value, used, allocated)                  \
{                                                                       \
    unsigned char *valuePtr;                                            \
                                                                        \
    if ((used) + 1 > (allocated)) {                                     \
        (allocated) += 10;                                              \
        (result) = realloc((result),                                    \
            sizeof(unsigned char *) * (allocated) + 1);                 \
    }                                                                   \
                                                                        \
    (result)[(used)] = strdup((value));                                 \
                                                                        \
    valuePtr = (result)[(used)];                                        \
    while (valuePtr[strlen(valuePtr)] == ' ') {                         \
        valuePtr[strlen(valuePtr)] = '\0';                              \
    }                                                                   \
                                                                        \
    valuePtr = FindMDBClass((result)[(used)]);                          \
    if (valuePtr) {                                                     \
        free((result)[(used)]);                                         \
        (result)[(used)] = strdup((valuePtr));                          \
    }                                                                   \
                                                                        \
    (used) += 1;                                                        \
    (result)[(used)] = NULL;                                            \
}

static __inline unsigned char **
GetSchemaValues(unsigned char *schema, unsigned char *key)
{
    unsigned char *k = strstr(schema, key);
    unsigned char **result = NULL;
    unsigned long used = 0;
    unsigned long allocated = 0;

    if (k) {
        unsigned char *s = k + strlen(key);
        unsigned char *e = NULL;
        unsigned char o;

        while (*s == ' ') {
            s++;
        }

        /* Find the end of this property. */
        switch (*s) {
            case '(':   s++;    o = ')';    break;
            case '\'':  s++;    o = '\'';   break;
            default:            o = ' ';    break;
        }

        e = strchr(s, o);
        if (e) {
            *e = '\0';

            /*
                Pull out each value

                The values will each be in single quotes seperated by a space, or seperated
                by a dollar sign.
            */
            while (*s == ' ') {
                s++;
            }

            if (*s == '\'') {
                unsigned char *m;

                do {
                    s++;

                    m = strchr(s, '\'');
                    if (m) {
                        *m = '\0';

                        addSchemaValue(result, s, used, allocated);

                        *m = '\'';
                        s = strchr(m + 1, '\'');
                    } else {
                        addSchemaValue(result, s, used, allocated);
                    }
                } while (s && m);
            } else {
                unsigned char *m;

                do {
                    while (*s == ' ') {
                        s++;
                    }

                    m = strchr(s, '$');
                    if (m) {
                        *m = '\0';

                        addSchemaValue(result, s, used, allocated);

                        *m = '$';
                        s = m + 1;
                    } else {
                        addSchemaValue(result, s, used, allocated);
                    }
                } while (m);
            }

            /* Restore the string */
            *e = o;
        }
    }

    return(result);
}

EXPORT BOOL 
MDBLDAPInit(MDBDriverInitStruct *Init)
{
    int i;

    /* allocate a bunch of these and pool them a little */
    AuthenticatedUsers = malloc(sizeof(MDBHandle *)*ALLOC_AUTHENTICATED);
    Alloc_Authenticated = ALLOC_AUTHENTICATED;
    AuthenticatedUsers[AuthenticatedCount++] = NULL;
    MDBldap.dirType = MDBLDAP_TYPE_NONE;
    MDBldap.treeName[0] = '\0';
    MDBldap.rootDSE[0]  = '\0';
    for (i = 0 ; i < (MDBLDAP_MAX_NAMING_CONTEXTS+1) ; i++) {
        MDBldap.namingContexts[i]  = NULL;
    }

    if (Init->Arguments) {
        MDBLDAPParseArguments(Init->Arguments);
    }

    /* open the outfile */
    if (outfilepath) {
        outfile = fopen(outfilepath, "w");
    }

    if (!outfile) {
        outfile = stdout;
    }

    Init->Interface.MDBAdd=MDBLDAPAdd;
    Init->Interface.MDBAddAddressRestriction=MDBLDAPAddAddressRestriction;
    Init->Interface.MDBAddAttribute=MDBLDAPAddAttribute;
    Init->Interface.MDBAddDN=MDBLDAPAddDN;
    Init->Interface.MDBAddValue=MDBLDAPAddValue;
    Init->Interface.MDBAuthenticate=MDBLDAPAuthenticate;
    Init->Interface.MDBChangePassword=MDBLDAPChangePassword;
    Init->Interface.MDBChangePasswordEx=MDBLDAPChangePasswordEx;
    Init->Interface.MDBClear=MDBLDAPClear;
    Init->Interface.MDBCreateEnumStruct=MDBLDAPCreateEnumStruct;
    Init->Interface.MDBCreateObject=MDBLDAPCreateObject;
    Init->Interface.MDBCreateValueStruct=MDBLDAPCreateValueStruct;
    Init->Interface.MDBDefineAttribute=MDBLDAPDefineAttribute;
    Init->Interface.MDBDefineClass=MDBLDAPDefineClass;
    Init->Interface.MDBDeleteObject=MDBLDAPDeleteObject;
    Init->Interface.MDBDestroyEnumStruct=MDBLDAPDestroyEnumStruct;
    Init->Interface.MDBDestroyValueStruct=MDBLDAPDestroyValueStruct;
    Init->Interface.MDBEnumerateAttributesEx=MDBLDAPEnumerateAttributesEx;
    Init->Interface.MDBEnumerateObjects=MDBLDAPEnumerateObjects;
    Init->Interface.MDBEnumerateObjectsEx=MDBLDAPEnumerateObjectsEx;
    Init->Interface.MDBFreeValue=MDBLDAPFreeValue;
    Init->Interface.MDBFreeValues=MDBLDAPFreeValues;
    Init->Interface.MDBGetObjectDetails=MDBLDAPGetObjectDetails;
    Init->Interface.MDBGetServerInfo=MDBLDAPGetServerInfo;
    Init->Interface.MDBGrantAttributeRights=MDBLDAPGrantAttributeRights;
    Init->Interface.MDBGrantObjectRights=MDBLDAPGrantObjectRights;
    Init->Interface.MDBIsObject=MDBLDAPIsObject;
    Init->Interface.MDBListContainableClasses=MDBLDAPListContainableClasses;
    Init->Interface.MDBListContainableClassesEx=MDBLDAPListContainableClassesEx;
    Init->Interface.MDBRead=MDBLDAPRead;
    Init->Interface.MDBReadDN=MDBLDAPReadDN;
    Init->Interface.MDBReadEx=MDBLDAPReadEx;
    Init->Interface.MDBRelease=MDBLDAPRelease;
    Init->Interface.MDBRemove=MDBLDAPRemove;
    Init->Interface.MDBRemoveDN=MDBLDAPRemoveDN;
    Init->Interface.MDBRenameObject=MDBLDAPRenameObject;
    Init->Interface.MDBSetValueStructContext=MDBLDAPSetValueStructContext;
    Init->Interface.MDBShareContext=MDBLDAPShareContext;
    Init->Interface.MDBUndefineAttribute=MDBLDAPUndefineAttribute;
    Init->Interface.MDBUndefineClass=MDBLDAPUndefineClass;
    Init->Interface.MDBVerifyPassword=MDBLDAPVerifyPassword;
    Init->Interface.MDBVerifyPasswordEx=MDBLDAPVerifyPassword;
    Init->Interface.MDBWrite=MDBLDAPWrite;
    Init->Interface.MDBWriteDN=MDBLDAPWriteDN;
    Init->Interface.MDBWriteTyped=MDBLDAPWriteTyped;

    /*  Authenticate as our admin so we can create the ncpServer object.     */
    MDBldap.mdbhandle = MDBLDAPAuthenticate(ldapUser, ldapPassword, "");
    
    if (ldapSchemaPath) {
#ifndef WIN32
        LoadTranslationTable(ldapSchemaPath);
#endif
    }
    
    /*  If the user specifies a cfgObject, we won't create the ncpServer obj */
    if (MDBldap.mdbhandle && !cfgObject) {
        char hostname[MDB_MAX_OBJECT_CHARS];
        int err;
        LDAPMod **attrs;
        LDAPMessage *pres = NULL;
        MDBValueStruct *v;

        gethostname(hostname, MDB_MAX_OBJECT_CHARS);
        sprintf(serverDN, "cn=%s,%s", hostname, ldapContext);

        XplWaitOnLocalSemaphore(MDBldap.mdbhandle->sem);
        err = ldap_search_s(MDBldap.mdbhandle->bindhandle, serverDN, LDAP_SCOPE_BASE, "(objectClass=*)", DNAttributes, 0, &pres);
        if (err!=LDAP_SUCCESS) {
            /* set up the array of attributes */
            attrs = malloc(3 * sizeof(LDAPMod *));
            attrs[0] = malloc(sizeof(LDAPMod));
            attrs[0]->mod_op = LDAP_MOD_BVALUES | LDAP_MOD_ADD;
            attrs[0]->mod_type = strdup("objectClass");
            attrs[0]->mod_bvalues = malloc(2*sizeof(struct berval *));
            attrs[0]->mod_bvalues[0] = malloc(sizeof(struct berval));
            attrs[0]->mod_bvalues[0]->bv_val = strdup("ncpServer");
            attrs[0]->mod_bvalues[0]->bv_len = strlen("ncpServer");
            attrs[0]->mod_bvalues[1] = NULL;

            attrs[1] = malloc(sizeof(LDAPMod));
            attrs[1]->mod_op = LDAP_MOD_BVALUES | LDAP_MOD_ADD;
            attrs[1]->mod_type = strdup("cn");
            attrs[1]->mod_bvalues = malloc(2*sizeof(struct berval *));
            attrs[1]->mod_bvalues[0] = malloc(sizeof(struct berval));
            attrs[1]->mod_bvalues[0]->bv_len = strlen(hostname);
            attrs[1]->mod_bvalues[0]->bv_val = strdup(hostname);
            attrs[1]->mod_bvalues[1] = NULL;

            attrs[2] = NULL;

            err = ldap_add_s(MDBldap.mdbhandle->bindhandle, serverDN, attrs);
            ldap_mods_free(attrs, 1);
        }
        if (pres) {
            ldap_msgfree(pres);
        }
        XplSignalLocalSemaphore(MDBldap.mdbhandle->sem);

        v = MDBLDAPCreateValueStruct(MDBldap.mdbhandle, NULL);
        if (v) {
            MDBLDAPFreeValues(v);

            /*  Read the schema into our structure for future lookups.       */
            if (MDBldap.subschemaSubentry[0] != '\0') {
                unsigned char *schema = MDBldap.subschemaSubentry;
                int err;
                LDAPMessage *pres = NULL;
                char *filter[2];
                char **values = NULL;

                filter[0] = "objectClasses";
                filter[1] = NULL;

                XplWaitOnLocalSemaphore(v->context->sem);

                /* Read the list of classes */
                err = ldap_search_s(v->context->bindhandle, schema, LDAP_SCOPE_BASE, "(objectClass=*)", filter, 0, &pres);
                if (err == LDAP_SUCCESS) {
                    LDAPMessage *pres2 = ldap_first_entry(v->context->bindhandle, pres);

                    if (pres2) {
                        values = ldap_get_values(v->context->bindhandle, pres2, filter[0]);
                        if (values) {
                            if ((i = ldap_count_values(values)) > 0) {
                                MDBldap.schema =
                                   malloc(sizeof(MDBldapClassStruct) * (i + 1));
                                   memset(MDBldap.schema, 0,
                                          sizeof(MDBldapClassStruct) * (i + 1));
                            }

                            i = 0;
                            while (values[i] != NULL) {
                                MDBldap.schema[i].names = GetSchemaValues(values[i], "NAME");
                                i++;
                            }

                            /* Get the sub classes */
                            i = 0;
                            while (values[i] != NULL) {
                                unsigned char **superclasses = GetSchemaValues(values[i], "SUP");

                                if (superclasses) {
                                    unsigned long s = 0;

                                    while (superclasses[s]) {
                                        MDBldapClassStruct *class = findClassDefinition(superclasses[s]);

                                        if (class) {
                                            /* Reallocate the classes sub classes list and add this to it */
                                            class->subClassesUsed += 1;
                                            class->subClasses = realloc(class->subClasses, sizeof(MDBldapClassStruct *) * (class->subClassesUsed + 1));

                                            class->subClasses[class->subClassesUsed - 1] = &MDBldap.schema[i];
                                            class->subClasses[class->subClassesUsed] = NULL;
                                        }

                                        s++;
                                    }

                                    free(superclasses);
                                }

                                i++;
                            }


                            /* Is it a container? */
                            i = 0;
                            while (values[i] != NULL) {
                                unsigned char **container = GetSchemaValues(values[i], "X-NDS_NOT_CONTAINER");

                                if (container) {
                                    unsigned long c = 0;

                                    while (container[c]) {
                                        if (container[c][0] == '1') {
                                            MDBldap.schema[i].leafOnly = TRUE;
                                            free(container[c]);
                                            c++;
                                        }
                                    }

                                    free(container);
                                }

                                i++;
                            }

                            /* Get the containment */
                            i = 0;
                            while (values[i] != NULL) {
                                unsigned char **containment = GetSchemaValues(values[i], "X-NDS_CONTAINMENT");

                                if (containment) {
                                    unsigned long c = 0;

                                    while (containment[c]) {
                                        MDBldapClassStruct *class = findClassDefinition(containment[c]);

                                        if (class) {
                                            /* Reallocate the classes containable classes list and add this to it */
                                            class->containableClassesUsed += 1;
                                            class->containableClasses = realloc(class->containableClasses, sizeof(MDBldapClassStruct *) * (class->containableClassesUsed + 1));

                                            class->containableClasses[class->containableClassesUsed - 1] = &MDBldap.schema[i];
                                            class->containableClasses[class->containableClassesUsed] = NULL;
                                        }

                                        c++;
                                    }

                                    free(containment);
                                } else {
                                    MDBldap.schema[i].inheritContainment = TRUE;
                                }

                                i++;
                            }

                            /* Get the naming */
                            i = 0;
                            while (values[i] != NULL) {
                                unsigned char **naming = GetSchemaValues(values[i], "X-NDS_NAMING");

                                if (naming) {
                                    MDBldap.schema[i].naming = naming;
                                    MDBldap.schema[i].freeNaming = TRUE;
                                } else {
                                    unsigned char **superclasses = GetSchemaValues(values[i], "SUP");

                                    if (superclasses) {
                                        unsigned long c = 0;
                                        unsigned long x = 0;
                                        /* Find a super class to copy it from */

                                        while (superclasses[x]) {
                                            while (MDBldap.schema[c].names) {
                                                unsigned long s = 0;

                                                while (MDBldap.schema[c].names[s]) {
                                                    if (MDBldap.schema[c].naming && strcmp(MDBldap.schema[c].names[s], superclasses[x]) == 0) {
                                                        MDBldap.schema[i].naming = MDBldap.schema[c].naming;
                                                        break;
                                                    }

                                                    s++;
                                                }
                                                c++;
                                            }
                                            x++;
                                        }

                                        free(superclasses);
                                    }
                                }

                                i++;
                            }

                            ldap_value_free(values);
                        }
                    }
                }

                if (pres) {
                    ldap_msgfree(pres);
                }

                XplSignalLocalSemaphore(v->context->sem);

                MDBLDAPFreeValues(v);
            }

            MDBLDAPDestroyValueStruct(v);
        }
    }

    return(TRUE);
}

EXPORT void 
MDBLDAPShutdown(void)
{
    int x=0;
    MDBHandle f;

    while ((f=AuthenticatedUsers[x])) {
        ldap_unbind_s(AuthenticatedUsers[x]->bindhandle);
        free(AuthenticatedUsers[x++]);
    }
    free(AuthenticatedUsers);

    if (ldapServer) {
        free(ldapServer);
    }

    if (ldapUser) {
        free(ldapUser);
    }

    if (ldapPassword) {
        free(ldapPassword);
    }

    if (ldapContext) {
        free(ldapContext);
    }

    if (ldapSchemaPath) {
        free(ldapSchemaPath);
    }

    if (cfgObject) {
        free(cfgObject);
    }

    fclose(outfile);
    if (outfilepath) {
        free(outfilepath);
    }

    /* Cleanup the schema */
    if (MDBldap.schema) {
        unsigned long i = 0;

        while (MDBldap.schema[i].names) {
            unsigned long n = 0;

            while (MDBldap.schema[i].names[n]) {
                free(MDBldap.schema[i].names[n]);
                n++;
            }
            

            if (MDBldap.schema[i].subClasses) {
                free(MDBldap.schema[i].subClasses);
            }

            if (MDBldap.schema[i].containableClasses) {
                free(MDBldap.schema[i].containableClasses);
            }

            if (MDBldap.schema[i].freeNaming && MDBldap.schema[i].naming) {
                free(MDBldap.schema[i].naming);
            }

            free(MDBldap.schema[i].names);
            i++;
        }

        free(MDBldap.schema);
    }

    for (x = 0 ; x < MDBLDAP_MAX_NAMING_CONTEXTS ; x++) {
        if (MDBldap.namingContexts[x] != NULL) {
            free(MDBldap.namingContexts[x]);
        } else {
            break;        /*  we've reached the end of our used list */
        }
    }

    MDBLDAPRelease(MDBldap.mdbhandle);
    return;
}
