#include <windows.h>
#include <stdio.h>

#include "ntsecapi.h"

NTSTATUS
OpenPolicy(
    LPWSTR ServerName,          // machine to open policy on (Unicode)
    DWORD DesiredAccess,        // desired access to policy
    PLSA_HANDLE PolicyHandle    // resultant policy handle
    );

BOOL
GetAccountSid(
    LPTSTR SystemName,          // where to lookup account
    LPTSTR AccountName,         // account of interest
    PSID *Sid                   // resultant buffer containing SID
    );

NTSTATUS
SetPrivilegeOnAccount(
    LSA_HANDLE PolicyHandle,    // open policy handle
    PSID AccountSid,            // SID to grant privilege to
    LPWSTR PrivilegeName,       // privilege to grant (Unicode)
    BOOL bEnable                // enable or disable
    );

void
InitLsaString(
    PLSA_UNICODE_STRING LsaString, // destination
    LPWSTR String                  // source (Unicode)
    );

#define RTN_OK 0
#define RTN_USAGE 1
#define RTN_ERROR 13

//
// If you have the ddk, include ntstatus.h.
//
#ifndef STATUS_SUCCESS
#define STATUS_SUCCESS  ((NTSTATUS)0x00000000L)
#endif


BOOL
GrantServiceRights(unsigned char *User)
{
	LSA_HANDLE PolicyHandle;
	PSID			pSid;
	NTSTATUS	Status;
	int			iRetVal=RTN_ERROR;          // assume error from main

	Status = OpenPolicy(L"", POLICY_CREATE_ACCOUNT | POLICY_LOOKUP_NAMES, &PolicyHandle);

	if(Status != STATUS_SUCCESS) {
	  return(FALSE);
	}

	//
	// Obtain the SID of the user/group.
	// Note that we could target a specific machine, but we don't.
	// Specifying NULL for target machine searches for the SID in the
	// following order: well-known, Built-in and local, primary domain,
	// trusted domains.
	//
	if(GetAccountSid(NULL, User, &pSid)) {
		Status = SetPrivilegeOnAccount(PolicyHandle, pSid, L"SeServiceLogonRight", TRUE);

		if(Status == STATUS_SUCCESS) {
			iRetVal = RTN_OK;
		}
	}

	LsaClose(PolicyHandle);

	if(pSid != NULL) {
		HeapFree(GetProcessHeap(), 0, pSid);
	}

	return(iRetVal);
}

BOOL
GetAccountSid(LPTSTR SystemName, LPTSTR AccountName, PSID *Sid)
{
	LPTSTR			ReferencedDomain=NULL;
	DWORD				cbSid=128;						// initial allocation attempt
	DWORD				cchReferencedDomain=16;		// initial allocation size
	SID_NAME_USE	peUse;
	BOOL				bSuccess=FALSE;				// assume this function will fail

	*Sid = (PSID)HeapAlloc(GetProcessHeap(), 0, cbSid);

	if(*Sid == NULL) {
		return(FALSE);
	}

	ReferencedDomain = (LPTSTR)HeapAlloc(GetProcessHeap(), 0, cchReferencedDomain * sizeof(TCHAR));

	if(ReferencedDomain == NULL) {
		if(*Sid != NULL) {
			HeapFree(GetProcessHeap(), 0, *Sid);
			*Sid = NULL;
		}
		return(FALSE);
	}

    //
    // Obtain the SID of the specified account on the specified system.
    //
	while(!LookupAccountName(SystemName, AccountName, *Sid, &cbSid, ReferencedDomain, &cchReferencedDomain, &peUse)) {
		if (GetLastError() == ERROR_INSUFFICIENT_BUFFER) {
			*Sid = (PSID)HeapReAlloc(GetProcessHeap(), 0, *Sid, cbSid);
			if(*Sid == NULL) {
				goto Shit;
			}

			ReferencedDomain = (LPTSTR)HeapReAlloc(GetProcessHeap(), 0, ReferencedDomain, cchReferencedDomain * sizeof(TCHAR));
			if(ReferencedDomain == NULL) {
				goto Shit;
			}
		}
	}

	bSuccess = TRUE;

    //
    // Cleanup and indicate failure, if appropriate.
    //
Shit:

	HeapFree(GetProcessHeap(), 0, ReferencedDomain);

	if(!bSuccess) {
		if(*Sid != NULL) {
			HeapFree(GetProcessHeap(), 0, *Sid);
			*Sid = NULL;
		}
	}

	return bSuccess;
}

NTSTATUS
SetPrivilegeOnAccount(LSA_HANDLE PolicyHandle, PSID AccountSid, LPWSTR PrivilegeName, BOOL bEnable)
{
	LSA_UNICODE_STRING PrivilegeString;

	InitLsaString(&PrivilegeString, PrivilegeName);

	if(bEnable) {
		return(LsaAddAccountRights(PolicyHandle, AccountSid, &PrivilegeString, 1));
	} else {
		return(LsaRemoveAccountRights(PolicyHandle, AccountSid, FALSE, &PrivilegeString, 1));
	}
}

void
InitLsaString(PLSA_UNICODE_STRING LsaString, LPWSTR String)
{
	DWORD StringLength;

	if(String == NULL) {
		LsaString->Buffer = NULL;
		LsaString->Length = 0;
		LsaString->MaximumLength = 0;
		return;
	}

	StringLength = lstrlenW(String);
	LsaString->Buffer = String;
	LsaString->Length = (USHORT) StringLength * sizeof(WCHAR);
	LsaString->MaximumLength=(USHORT)(StringLength+1) * sizeof(WCHAR);
}

NTSTATUS
OpenPolicy(LPWSTR ServerName, DWORD DesiredAccess, PLSA_HANDLE PolicyHandle)
{
    LSA_OBJECT_ATTRIBUTES	ObjectAttributes;
    LSA_UNICODE_STRING		ServerString;
    PLSA_UNICODE_STRING		Server;

    ZeroMemory(&ObjectAttributes, sizeof(ObjectAttributes));

    if (ServerName != NULL) {
        InitLsaString(&ServerString, ServerName);
        Server = &ServerString;
    } else {
        Server = NULL;
    }

    return(LsaOpenPolicy(Server, &ObjectAttributes, DesiredAccess, PolicyHandle));
}

